#!/usr/bin/env python3

import os
from math import sqrt

__NAME__ = "Sort_paf"
__VERSION__ = 0.1


class Sorter:
    """
    Sort PAF file by match size
    """

    def __init__(self, input_f, output_f):
        """

        :param input_f: input fasta file path
        :type input_f: str
        :param output_f: output fasta file path
        :type output_f: str
        """
        self.input_f = input_f
        self.output_f = output_f

    def sort(self):
        """
        Launch sort staff
        """
        paf_lines = self._get_sorted_paf_lines()
        with open(self.output_f, "w") as out:
            out.write("\n".join(["\t".join(x[:-1]) for x in paf_lines]))

    def _sort_lines(self, lines):
        """
        Sort lines staff

        :param lines: lines of PAF file to be sorted
        :type lines: _io.TextIO
        :return: sorted lines
        :rtype: list
        """
        paf_lines = []
        nb_lines = 0
        min_len = 0
        for line in lines:
            parts = line.strip("\n").split("\t")
            len_line = sqrt(pow(int(parts[3]) - int(parts[2]), 2) +
                            pow(int(parts[8]) - int(parts[7]), 2)) * (int(parts[9]) / int(parts[10]))
            parts.append(len_line)
            nb_lines += 1
            if nb_lines <= 1000000:
                paf_lines.append(parts)
                if nb_lines == 1000000:
                    print("Sorting lines...")
                    paf_lines.sort(key=lambda x: -x[-1])
                    with open(self.input_f, 'r') as paf:
                        num_lines = sum(1 for line in paf)
                    pct_to_keep = 2000000 / num_lines
                    limit = int(nb_lines * pct_to_keep)
                    paf_lines = paf_lines[:limit]
                    min_len = paf_lines[-1][-1]
                    print("Sorting done!")
            elif len_line >= min_len:
                print("Processing line %d..." % nb_lines)
                paf_lines.append(parts)
        paf_lines.sort(key=lambda x: -x[-1])

        return paf_lines

    def _get_sorted_paf_lines(self):
        """
        Get sorted PAF

        :return: sorted PAF lines
        """
        with open(self.input_f, "r") as lines:
            paf_lines = self._sort_lines(lines)
        return paf_lines


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description="Sort PAF file with size of matches (DESC)")
    parser.add_argument('-i', '--input', type=str, required=True, help="Input PAF file")
    parser.add_argument('-o', '--output', type=str, required=True, help="Output PAF file")

    args = parser.parse_args()
    if args["--version"]:
        print(__NAME__, __VERSION__)
    else:
        if not os.path.exists(args.input):
            raise Exception("Input PAF file %s does not exists" % args.input)
        sorter = Sorter(args.input, args.output)
        sorter.sort()



#!/usr/bin/env python3

import os
import argparse
import tempfile
from lib.parse_paf import *
import prepare_paf

parser = argparse.ArgumentParser(description="Convert paf file to json file readable by algeco tool")
parser.add_argument('-i', '--input', type=str, required=True, help="Paf file to parse")
parser.add_argument('-q', '--query', type=str, required=True, help="Query fasta file")
parser.add_argument('-t', '--target', type=str, required=True, help="Target fasta file")
parser.add_argument('-o', '--output', type=str, required=True, help="Output json file")

args = parser.parse_args()

tmp_dir = tempfile.mkdtemp()
paf_out = os.path.join(tmp_dir, "map.paf")

# Prepare paf file
prepare_paf.init(args.input, paf_out, args.query, args.target)

# Build json
idx1 = os.path.join(tmp_dir, "query.idx")
idx2 = os.path.join(tmp_dir, "target.idx")
save_json(paf_out, idx1, idx2, args.output)



#!/usr/bin/env python3

import re
from collections import OrderedDict


class Merger:

    """
    Merge splitted contigs together in PAF file
    """

    def __init__(self, paf_in, paf_out, query_in, query_out, debug=False):
        """

        :param paf_in: input PAF file path
        :type paf_in: str
        :param paf_out: output PAF file path
        :type paf_out: str
        :param query_in: input query index file path
        :type query_in: str
        :param query_out: output query index file path
        :type query_out: str
        :param debug: True to enable debug mode
        :type debug: bool
        """
        self.paf_in = paf_in
        self.paf_out = paf_out
        self.query_in = query_in
        self.query_out = query_out
        self.debug = debug

    def _printer(self, message):
        """
        Print debug messages if debug mode enabled

        :param message: message to print
        :type message: str
        """
        if self.debug:
            print(message)

    def merge(self):
        """
        Launch the merge
        """
        self._printer("Loading query index...")
        contigs, contigs_split, q_name = self.load_query_index(self.query_in)

        self._printer("Merging contigs in PAF file...")
        self.merge_paf(self.paf_in, self.paf_out, contigs, contigs_split)

        self._printer("Writing new query index...")
        self.write_query_index(self.query_out, contigs, q_name)

        self._printer("DONE!")

    @staticmethod
    def _get_sorted_splits(contigs_split, all_contigs):
        """
        For each contigs_split, save how many base we will must add to each line of the corresponding split contig in PAF
        file.
        Also, save the final merged contig size in all contig dict

        :param contigs_split: split contigs
        :type contigs_split: dict
        :param all_contigs: all and final contigs
        :type all_contigs: dict
        :return: all contigs and new split contigs with start of each split contig set
        :rtype: (dict, dict)
        """
        new_contigs = {}
        for contig, splits_d in contigs_split.items():
            new_contigs[contig] = OrderedDict()
            splits = sorted(list(splits_d.keys()), key=lambda x: int(x))
            cum_len = 0
            for split in splits:
                new_contigs[contig][split] = cum_len  # What must be added to each line in this contig
                cum_len += splits_d[split]
            all_contigs[contig] = cum_len
        return all_contigs, new_contigs

    def load_query_index(self, index):
        """
        Load query index

        :param index: index file path
        :type index: str
        :return:
            * [0] contigs length
            * [1] splitted contigs length
            * [2] sample name
        :rtype: (dict, dict, str)
        """
        contigs = OrderedDict()
        contigs_split = {}
        with open(index) as idx_f:
            q_name = idx_f.readline().strip("\n")
            for line in idx_f:
                parts = line.strip("\n").split("\t")
                contig_name = parts[0]
                contig_len = int(parts[1])
                match_split = re.match(r"^(.+)_###_(\d+)$", contig_name)
                if match_split is not None:
                    contig_name = match_split.group(1)
                    if contig_name not in contigs_split:
                        contigs_split[contig_name] = {}
                    nb_split = match_split.group(2)
                    contigs_split[contig_name][nb_split] = contig_len
                    contigs[contig_name] = None  # Will be filled after
                else:
                    contigs[contig_name] = contig_len
        if len(contigs_split) > 0:
            contigs, contigs_split = self._get_sorted_splits(contigs_split, contigs)
        return contigs, contigs_split, q_name


    @staticmethod
    def write_query_index(index, contigs, q_name):
        """
        Save new query index

        :param index: index file path
        :type index: str
        :param contigs: contigs size
        :type contigs: dict
        :param q_name: sample name
        :type q_name: str
        """
        with open(index, "w") as idx_f:
            idx_f.write(q_name + "\n")
            for contig_name, contig_len in contigs.items():
                idx_f.write("%s\t%d\n" % (contig_name, contig_len))

    @staticmethod
    def merge_paf(paf_in, paf_out, contigs, contigs_split):
        """
        Do merge PAF staff

        :param paf_in: path of input PAF with split contigs
        :type paf_in: str
        :param paf_out: path of output PAF where split contigs are now merged together
        :type paf_out: str
        :param contigs: contigs size
        :type contigs: dict
        :param contigs_split: split contigs size
        :type contigs_split: dict
        """
        with open(paf_in) as paf_i, open(paf_out, "w") as paf_o:
            for line in paf_i:
                parts = line.strip("\n").split("\t")
                match_split = re.match(r"^(.+)_###_(\d+)$", parts[0])
                if match_split is None:
                    paf_o.write(line)
                else:
                    contig_name = match_split.group(1)
                    nb_split = match_split.group(2)
                    parts[0] = contig_name
                    parts[1] = str(contigs[contig_name])
                    parts[2] = str(int(parts[2]) + contigs_split[contig_name][nb_split])
                    parts[3] = str(int(parts[3]) + contigs_split[contig_name][nb_split])
                    paf_o.write("\t".join(parts) + "\n")


def parse_args():
    """
    Parse command line arguments

    :return: arguments
    :rtype: argparse.Namespace
    """
    import argparse

    parser = argparse.ArgumentParser(description='Merge in PAF file and indexed when fasta has been split',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-pi", "--paf-in", help="Input PAF file", type=str, required=True)
    parser.add_argument("-po", "--paf-out", help="Output PAF file", type=str, required=True)
    parser.add_argument("-qi", "--query-in", help="Input query index file", type=str, required=True)
    parser.add_argument("-qo", "--query-out", help="Output query index file", type=str, required=True)
    p_args = parser.parse_args()
    return p_args


if __name__ == '__main__':
    args = parse_args()
    merger = Merger(args.paf_in, args.paf_out, args.query_in, args.query_out)
    exit(merger.merge())



class UploadFile:
    """
    Manage uploaded files
    """

    def __init__(self, name, type_f=None, size=None, not_allowed_msg=''):
        """

        :param name: File name
        :type name: str
        :param type_f: file MIME type
        :type type_f: str
        :param size: file size in bytes
        :type size: int
        :param not_allowed_msg: error to add for not allowed file
        :type not_allowed_msg: str
        """
        self.name = name
        self.type = type_f
        self.size = size
        self.not_allowed_msg = not_allowed_msg
        self.url = "data/%s" % name

    def get_file(self):
        """
        Get file object

        :return: file object
        :rtype: dict
        """
        if self.type is not None:
            # POST an image
            if self.type.startswith('image'):
                return {"name": self.name,
                        "type": self.type,
                        "size": self.size, 
                        "url": self.url}
            
            # POST a normal file
            elif self.not_allowed_msg == '':
                return {"name": self.name,
                        "type": self.type,
                        "size": self.size, 
                        "url": self.url}

            # File type is not allowed
            else:
                return {"error": self.not_allowed_msg,
                        "name": self.name,
                        "type": self.type,
                        "size": self.size}
        
        # GET normal file from disk
        else:
            return {"name": self.name,
                    "size": self.size, 
                    "url": self.url}



"""
Define tools parsers here

Each parser (main function) must have 2 and only 2 arguments:
- First argument: input file which is the tool raw output
- Second argument: finale PAF file

Returns True if parse succeed, else False
"""

import traceback
from Bio import AlignIO


def maf(in_maf, out_paf):
    """
    Maf parser

    :param in_maf: input maf file path
    :type in_maf: str
    :param out_paf:  output paf file path
    :type out_paf: str
    :return: True if success, else False
    """
    maf_f = None
    try:
        with open(out_paf, "w") as paf:
            maf_f = AlignIO.parse(in_maf, "maf")
            for grp in maf_f:
                seqs = []
                for seq in grp:
                    seqs.append(seq)
                matches = 0
                for i in range(0, len(seqs[0])):
                    if seqs[0][i] == seqs[1][i]:
                        matches += 1
                tannots = seqs[0].annotations
                qannots = seqs[1].annotations
                tlen = tannots["srcSize"]
                tstart = tannots["start"]
                tend = tstart + tannots["size"]
                if tannots["strand"] == -1:
                    tstart = tlen - tstart
                    tend = tlen - tend
                qlen = qannots["srcSize"]
                qstart = qannots["start"]
                qend = qannots["start"] + qannots["size"]
                if qannots["strand"] == -1:
                    qstart = qlen - qstart
                    qend = qlen - qend
                strand = "+" if tannots["strand"] == qannots["strand"] else "-"
                paf.write("{qname}\t{qlen}\t{qstart}\t{qend}\t{strand}\t{tname}\t{tlen}\t{tstart}\t{tend}\t{matches}\t"
                          "{block_len}\t255\n".format(
                            tname=seqs[0].id,
                            tlen=tlen,
                            tstart=tstart,
                            tend=tend,
                            qname=seqs[1].id,
                            qlen=qlen,
                            qstart=qstart if strand == "+" else qend,
                            qend=qend if strand == "+" else qstart,
                            strand=strand,
                            matches=matches,
                            block_len=tannots["size"]
                          ))
    except:
        traceback.print_exc()
        if maf_f is not None:
            maf_f.close()
        return False
    else:
        maf_f.close()
        return True


def mashmap2paf(in_paf, out_paf):
    with open(in_paf, "r") as in_p, open(out_paf, "w") as out_p:
        for line in in_p:
            parts = line.rstrip().split(" ")
            parts[9] = str(round(float(parts[9]) / 100.0 * 1000.0))
            parts.append("1000")
            out_p.write("\t".join(parts) + "\t255" + "\n")




#!/usr/bin/env python3

import io
import re
from xopen import xopen


class Index:

    """
    Manage Fasta Index
    """

    def __init__(self):
        pass

    @staticmethod
    def load(index_file, merge_splits=False):
        """
        Load index

        :param index_file: index file path
        :type index: str
        :param merge_splits: if True, merge split contigs together
        :type merge_splits: bool
        :return:
            * [0] sample name
            * [1] contigs order
            * [2] contigs size
            * [3] reversed status for each contig
            * [4] absolute start position for each contig
            * [5] total len of the sample
        :rtype: (str, list, dict, dict, dict, int)
        """
        with open(index_file, "r") as idx_q_f:
            abs_start = {}
            abs_current_start = 0
            c_len = 0
            name = idx_q_f.readline().strip("\n")
            order = []
            contigs = {}
            reversed_c = {}
            for line in idx_q_f:
                parts = line.strip("\n").split("\t")
                id_c = parts[0]
                is_split = False
                if merge_splits:
                    match = re.match(r"(.+)_###_\d+", id_c)
                    if match is not None:
                        id_c = match.group(1)
                        is_split = True
                len_c = int(parts[1])
                if len(parts) > 2:
                    reversed_c[id_c] = parts[2] == "1"
                else:
                    reversed_c[id_c] = False
                if not is_split or (is_split and id_c not in order):
                    order.append(id_c)
                    abs_start[id_c] = abs_current_start
                    contigs[id_c] = len_c
                else:
                    contigs[id_c] += len_c
                c_len += len_c
                abs_current_start += len_c
            return name, order, contigs, reversed_c, abs_start, c_len

    @staticmethod
    def save(index_file, name, contigs, order, reversed_c):
        """
        Save index

        :param index_file: index file path
        :type index_file: str
        :param name: sample name
        :type name: str
        :param contigs: contigs size
        :type contigs: dict
        :param order: contifs order
        :type order: list
        :param reversed_c: reversed status for each contig
        :type reversed_c: dict
        """
        with open(index_file, "w") as idx:
            idx.write(name + "\n")
            for contig in order:
                idx.write("\t".join([contig, str(contigs[contig]), "1" if reversed_c[contig] else "0"])
                          + "\n")


def index_file(fasta_path, fasta_name, out, write_fa=None):
    """
    Index fasta file

    :param fasta_path: fasta file path
    :type fasta_path: str
    :param fasta_name: sample name
    :type fasta_name: str
    :param out: output index file
    :type out: str
    :param write_fa: file path of the new fasta file to write, None to don't save fasta in a new file
    :type write_fa: str
    :return:
        * [0] True if success, else False
        * [1] Number of contigs
        * [2] Error message
    :rtype: (bool, int, str)
    """
    has_header = False
    next_header = False  # True if next line must be a header line
    compressed = fasta_path.endswith(".gz")
    nb_contigs = 0
    write_f = None
    if write_fa is not None:
        write_f = open(write_fa, "w")
    with (xopen(fasta_path) if compressed else open(fasta_path)) as fasta, \
            open(out, "w") as out_file:
        out_file.write(fasta_name + "\n")
        contig = None
        len_c = 0
        nb_line = 0
        for line in fasta:
            nb_line += 1
            if write_f is not None:
                write_f.write(line)
            line = line.strip("\n")
            if re.match(r"^>.+", line) is not None:
                has_header = True
                next_header = False
                if contig is not None:
                    if len_c > 0:
                        nb_contigs += 1
                        out_file.write("%s\t%d\n" % (contig, len_c))
                    else:
                        return False, 0, "Error: contig is empty: %s" % contig
                contig = re.split("\s", line[1:])[0]
                len_c = 0
            elif len(line) > 0:
                if next_header or re.match(r"^[ATGCKMRYSWBVHDXN.\-]+$", line.upper()) is None:
                    if next_header:
                        return False, 0, "Error: new header line expected at line %d" % nb_line
                    return False, 0, "Error: invalid sequence at line %d" % nb_line
                len_c += len(line)
            elif len(line) == 0:
                next_header = True

        if contig is not None and len_c > 0:
            nb_contigs += 1
            out_file.write("%s\t%d\n" % (contig, len_c))

    if write_f is not None:
        write_f.close()

    return has_header, nb_contigs, ""


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description="Split huge contigs")
    parser.add_argument('-i', '--input', type=str, required=True, help="Input fasta file")
    parser.add_argument('-n', '--name', type=str, required=True, help="Input name")
    parser.add_argument('-o', '--output', type=str, required=True, help="Output index file")
    args = parser.parse_args()

    success, nb_contigs, message = index_file(args.input, args.name, args.output)
    if success:
        print("Success!")
    else:
        print(message)



from dgenies.config_reader import AppConfigReader
from flask_mail import Mail, Message


class Mailer:

    """
    Send mail throw flask app
    """

    def __init__(self, app):
        """

        :param app: Flask app object
        :type app: Flask
        """
        self.app = app
        self.mail = Mail(app)
        self.config = AppConfigReader()
        # self.mail_status = config_reader.get_mail_status_sender()
        # self.mail_reply = config_reader.get_mail_reply()
        # self.mail_org = config_reader.get_mail_org()
        # self.disable = config_reader.get_disable_mail()

    def _send_async_email(self, msg):
        """
        Send mail asynchronously

        :param msg: message to send
        :type msg: Message
        """
        with self.app.app_context():
            self.mail.send(msg)

    def send_mail(self, recipients, subject, message, message_html=None):
        """
        Send mail

        :param recipients: list of recipients
        :type recipients: list
        :param subject: mail subject
        :type subject: str
        :param message: message (text)
        :type message: str
        :param message_html: message (html)
        :type message_html: str
        """
        sender = (self.config.mail_org, self.config.mail_status_sender) if self.config.mail_org is not None else \
            self.config.mail_status_sender
        reply = self.config.mail_reply
        if not self.config.disable_mail:
            msg = Message(
                subject= subject,
                recipients=recipients,
                html=message_html,
                body=message,
                sender=sender,
                reply_to=reply
            )
            self._send_async_email(msg)
        else:  # Print debug
            print("################\n"
                  "# WARNING !!!! #\n"
                  "################\n\n"
                  "!!! SEND MAILS DISABLED BY CONFIGURATION !!!\n\n"
                  "(This might be disabled in production)\n\n")
            print("Sender: %s <%s>\n" % sender if isinstance(sender, tuple) else ("None", sender))
            print("Reply to: %s\n" % reply)
            print("Recipients: %s\n" % ", ".join(recipients))
            print("Subject: %s\n" % subject)
            print("Message:\n%s\n\n" % message)
            print("Message HTML:\n%s\n\n" % message_html)



from dgenies import MODE, DEBUG

import os
import shutil
import subprocess
from datetime import datetime
import time
import threading
import re
from dgenies.config_reader import AppConfigReader
from dgenies.tools import Tools
import dgenies.lib.validators as validators
import dgenies.lib.parsers as parsers
from .fasta import Fasta
from .functions import Functions
import requests
from requests.exceptions import ConnectionError
from urllib.request import urlretrieve
from urllib.error import URLError
from jinja2 import Template
import traceback
from pathlib import Path
from urllib import request, parse
import tarfile
from dgenies.bin.split_fa import Splitter
from dgenies.bin.index import index_file, Index
from dgenies.bin.filter_contigs import Filter
from dgenies.bin.merge_splitted_chrms import Merger
from dgenies.bin.sort_paf import Sorter
from dgenies.lib.paf import Paf
import gzip
import io
import binascii
from dgenies.database import Job

if MODE == "webserver":
    from dgenies.database import Session, Gallery
    from peewee import DoesNotExist


class JobManager:
    """
    Jobs management
    """

    def __init__(self, id_job, email=None, query: Fasta=None, target: Fasta=None, mailer=None,
                 tool="minimap2", align: Fasta=None, backup: Fasta=None, options=None):
        """
        :param id_job: job id
        :type id_job: str
        :param email: email from user
        :type email: str
        :param query: query fasta
        :type query: Fasta
        :param target: target fasta
        :type target: Fasta
        :param mailer: mailer object (to send mail throw flask app)
        :type mailer: Mailer
        :param tool: tool to use for mapping (choice from tools config)
        :type tool: str
        :param align: alignment file (PAF, MAF, ...) as a fasta object
        :type align: Fasta
        :param backup: backup TAR file
        :type backup: Fasta
        :param options: list of str containing options for the chosen tool
        :type options: list
        """
        self.id_job = id_job
        self.email = email
        self.query = query
        self.target = target
        self.align = align
        if align is not None:
            self.aln_format = os.path.splitext(align.get_path())[1][1:]
        self.backup = backup
        self.error = ""
        self.id_process = "-1"
        # Get configs:
        self.config = AppConfigReader()
        self.tools = Tools().tools
        self.tool = self.tools[tool] if tool is not None else None
        self.tool_name = tool
        self.options = options if tool is not None or options is not None else None
        # Outputs:
        self.output_dir = os.path.join(self.config.app_data, id_job)
        self.preptime_file = os.path.join(self.output_dir, "prep_times")
        self.query_index_split = os.path.join(self.output_dir, "query_split.idx")
        self.paf = os.path.join(self.output_dir, "map.paf")
        self.paf_raw = os.path.join(self.output_dir, "map_raw.paf")
        self.idx_q = os.path.join(self.output_dir, "query.idx")
        self.idx_t = os.path.join(self.output_dir, "target.idx")
        self.logs = os.path.join(self.output_dir, "logs.txt")
        self.mailer = mailer
        self._filename_for_url = {}

    def do_align(self):
        """
        Check if we have to make alignment

        :return: True if the job is launched with an alignment file
        """
        return not os.path.exists(os.path.join(self.output_dir, ".align"))

    @staticmethod
    def is_gz_file(filepath):
        """
        Check if a file is gzipped

        :param filepath: file to check
        :type filepath: str
        :return: True if gzipped, else False
        """
        with open(filepath, 'rb') as test_f:
            return binascii.hexlify(test_f.read(2)) == b'1f8b'

    def get_file_size(self, filepath: str):
        """
        Get file size

        :param filepath: file path
        :type filepath: str
        :return: file size (bytes)
        :rtype: int
        """
        file_size = os.path.getsize(filepath)
        if filepath.endswith(".gz") and file_size <= self.config.max_upload_size:
            with gzip.open(filepath, 'rb') as file_obj:
                file_size = file_obj.seek(0, io.SEEK_END)
        return file_size

    def get_query_split(self):
        """
        Get query split fasta file

        :return: split query fasta file
        :rtype: str
        """
        if not self.tool.split_before:
            return self.query.get_path()
        query_split = os.path.join(self.output_dir, "split_" + os.path.basename(self.query.get_path()))
        if query_split.endswith(".gz"):
            return query_split[:-3]
        return query_split

    def set_inputs_from_res_dir(self):
        """
        Sets inputs (query, target, ...) from job dir
        """
        res_dir = os.path.join(self.config.app_data, self.id_job)
        query_file = os.path.join(res_dir, ".query")
        if os.path.exists(query_file):
            with open(query_file) as q_f:
                file_path = q_f.readline()
                self.query = Fasta(
                    name="target" if file_path.endswith(".idx") else
                         os.path.splitext(os.path.basename(file_path.replace(".gz", "")).split("_", 1)[1])[0],
                    path=file_path,
                    type_f="local"
                )
        target_file = os.path.join(res_dir, ".target")
        if os.path.exists(target_file):
            with open(target_file) as t_f:
                file_path = t_f.readline()
                self.target = Fasta(
                    name="query" if file_path.endswith(".idx") else
                         os.path.splitext(os.path.basename(file_path.replace(".gz", "")).split("_", 1)[1])[0],
                    path=file_path,
                    type_f="local"
                )
        align_file = os.path.join(res_dir, ".align")
        if os.path.exists(align_file):
            with open(align_file) as a_f:
                file_path = a_f.readline()
                self.align = Fasta(
                    name="map",
                    path=file_path,
                    type_f="local"
                )
                self.aln_format = os.path.splitext(file_path)[1][1:]

    def check_job_success(self):
        """
        Check if a job succeed

        :return: status of a job: succeed, no-match or fail
        :rtype: str
        """
        if os.path.exists(self.paf_raw):
            if os.path.getsize(self.paf_raw) > 0:
                return "succeed"
            else:
                return "no-match"
        return "fail"

    def is_query_filtered(self):
        """
        Check if query has been filtered

        :return: True if filtered, else False
        """
        return os.path.exists(os.path.join(self.output_dir, ".filter-query"))

    def is_target_filtered(self):
        """
        Check if target has been filtered

        :return: True if filtered, else False
        :return:
        """
        return os.path.exists(os.path.join(self.output_dir, ".filter-target"))

    def get_mail_content(self, status, target_name, query_name=None):
        """
        Build mail content for status mail

        :param status: job status
        :type status: str
        :param target_name: name of target
        :type target_name: str
        :param query_name:  name of query
        :type query_name: str
        :return: mail content
        :rtype: str
        """
        message = "D-Genies\n\n"
        if status == "success":
            message += "Your job %s was completed successfully!\n\n" % self.id_job
            message += str("Your job {0} is finished. You can see  the results by clicking on the link below:\n"
                           "{1}/result/{0}\n\n").format(self.id_job, self.config.web_url)
        else:
            message += "Your job %s has failed!\n\n" % self.id_job
            if self.error != "":
                message += self.error.replace("#ID#", self.id_job).replace("<br/>", "\n")
                message += "\n\n"
            else:
                message += "Your job %s has failed. You can try again. " \
                           "If the problem persists, please contact the support.\n\n" % self.id_job
        if target_name is not None:
            message += "Sequences compared in this analysis:\n"
            if query_name is not None:
                message += "Target: %s\nQuery: %s\n\n" % (target_name, query_name)
            else:
                message += "Target: %s\n\n" % target_name
        if status == "success":
            if self.is_target_filtered():
                message += str("Note: target fasta has been filtered because it contains too small contigs."
                               "To see which contigs has been removed from the analysis, click on the link below:\n"
                               "{1}/filter-out/{0}/target\n\n").format(self.id_job, self.config.web_url)
            if self.is_query_filtered():
                message += str("Note: query fasta has been filtered because it contains too small contigs."
                               "To see which contigs has been removed from the analysis, click on the link below:\n"
                               "{1}/filter-out/{0}/query\n\n").format(self.id_job, self.config.web_url)
        message += "See you soon on D-Genies,\n"
        message += "The team"
        return message

    def get_mail_content_html(self, status, target_name, query_name=None):
        """
        Build mail content as HTML

        :param status: job status
        :type status: str
        :param target_name: name of target
        :type target_name: str
        :param query_name:  name of query
        :type query_name: str
        :return: mail content (html)
        :rtype: str
        """
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "mail_templates", "job_notification.html"))\
                as t_file:
            template = Template(t_file.read())
            return template.render(job_name=self.id_job, status=status, url_base=self.config.web_url,
                                   query_name=query_name if query_name is not None else "",
                                   target_name=target_name if target_name is not None else "",
                                   error=self.error,
                                   target_filtered=self.is_target_filtered(), query_filtered=self.is_query_filtered())

    def get_mail_subject(self, status):
        """
        Build mail subject

        :param status: job status
        :type status: str
        :return: mail subject
        :rtype: str
        """

        if status == "success" or status == "no-match":
            return "DGenies - Job completed: %s" % self.id_job
        else:
            return "DGenies - Job failed: %s" % self.id_job

    def send_mail(self):
        """
        Send mail
        """
        # Retrieve infos:
        with Job.connect():
            job = Job.get(Job.id_job == self.id_job)
            if self.email is None:
                self.email = job.email
            status = job.status
            self.error = job.error

            target_name = None
            if os.path.exists(self.idx_t):
                with open(self.idx_t, "r") as idxt:
                    target_name = idxt.readline().rstrip()
            query_name = None
            if os.path.exists(self.idx_q):
                with open(self.idx_q, "r") as idxq:
                    query_name = idxq.readline().rstrip()
                    if query_name == target_name:
                        query_name = None

            # Send:
            self.mailer.send_mail(recipients=[self.email],
                                  subject=self.get_mail_subject(status),
                                  message=self.get_mail_content(status, target_name, query_name),
                                  message_html=self.get_mail_content_html(status, target_name, query_name))

    def search_error(self):
        """
        Search for an error in the log file (for local runs). If no error found, returns a generic error message

        :return: error message to give to the user
        :rtype: str
        """
        logs = os.path.join(self.output_dir, "logs.txt")
        if os.path.exists(logs) and os.name == 'posix':
            lines = subprocess.check_output(['tail', '-2', logs]).decode("utf-8").split("\n")
            if re.match(r"\[morecore] \d+ bytes requested but not available.", lines[1]) or \
                    re.match(r"\[morecore] \d+ bytes requested but not available.", lines[1]) or \
                    re.match(r"\[morecore] insufficient memory", lines[0]) or \
                    re.match(r"\[morecore] insufficient memory", lines[1]):
                return "Your job #ID# has failed because of memory limit exceeded. May be your sequences are too big?" \
                       "<br/>You can contact the support for more information."
        return "Your job #ID# has failed. You can try again.<br/>If the problem persists, please contact the support."

    def _launch_local(self):
        """
        Launch a job on the current machine

        :return: True if job succeed, else False
        :rtype: bool
        """
        if MODE == "webserver":
            cmd = ["/usr/bin/time", "-f", "%e %M"]
        else:
            cmd = []
        if self.query is not None:
            command_line = self.tool.command_line.replace("{query}", self.query.get_path())
        else:
            command_line = self.tool.all_vs_all
        out_file = None
        if ">" in command_line:
            out_file = self.paf_raw
            command_line = command_line[:command_line.index(">")]
        command_line = command_line.replace("{exe}", self.tool.exec) \
                                   .replace("{target}", self.target.get_path()) \
                                   .replace("{threads}", str(self.tool.threads)) \
                                   .replace("{options}", self.options) \
                                   .replace("{out}", self.paf_raw)

        cmd += command_line.split(" ")
        if out_file is None:
            with open(self.logs, "w") as logs:
                p = subprocess.Popen(cmd, stdout=logs, stderr=logs)
        else:
            with open(self.logs, "w") as logs, open(out_file, "w") as out:
                p = subprocess.Popen(cmd, stdout=out, stderr=logs)
        with Job.connect():
            status = "started"
            if MODE == "webserver":
                job = Job.get(Job.id_job == self.id_job)
                job.id_process = p.pid
                job.status = status
                job.save()
            else:
                job = None
                self.set_status_standalone(status)
            p.wait()
            if p.returncode == 0:
                status = self.check_job_success()
                if MODE == "webserver":
                    job.status = status
                    job.save()
                else:
                    self.set_status_standalone(status)
                if status == "no-match":
                    self._set_analytics_job_status("no-match")
                return status == "succeed"
            self.error = self.search_error()
            status = "fail"
            if MODE == "webserver":
                job.status = status
                job.error = self.error
                job.save()
            else:
                self.set_status_standalone(status, self.error)
        return False

    def check_job_status_slurm(self):
        """
        Check status of a SLURM job run

        :return: True if the job has successfully ended, else False
        """
        status = subprocess.check_output("sacct -p -n --format=state,maxvmsize,elapsed -j %s.batch" % self.id_process,
                                         shell=True).decode("utf-8").strip("\n")

        status = status.split("|")

        success = status[0] == "COMPLETED"
        if success:
            mem_peak = int(status[1][:-1])  # Remove the K letter
            elapsed_full = list(map(int, status[2].split(":")))
            elapsed = elapsed_full[0] * 3600 + elapsed_full[1] * 60 + elapsed_full[2]
            with open(self.logs, "a") as logs:
                logs.write("%s %d" % (elapsed, mem_peak))

        return success

    def check_job_status_sge(self):
        """
        Check status of a SGE job run

        :return: True if the job jas successfully ended, else False
        """
        status = "-1"
        start = None
        end = None
        mem_peak = None
        acct = subprocess.check_output("qacct -d 1 -j %s" % self.id_process,
                                         shell=True).decode("utf-8")
        lines = acct.split("\n")
        for line in lines:
            if line.startswith("failed"):
                status = re.split(r"\s+", line, 1)[1]
            elif line.startswith("start_time"):
                start = datetime.strptime(re.split(r"\s+", line, 1)[1], "%a %b %d %H:%M:%S %Y")
            elif line.startswith("end_time"):
                end = datetime.strptime(re.split(r"\s+", line, 1)[1], "%a %b %d %H:%M:%S %Y")
            elif line.startswith("maxvmem"):
                mem_peak = re.split(r"\s+", line, 1)[1]
                if mem_peak.endswith("G"):
                    mem_peak = int(mem_peak[-1]) * 1024 * 1024
                elif mem_peak.endswith("M"):
                    mem_peak = int(mem_peak[-1]) * 1024

        if status == "0":
            if start is not None and end is not None and mem_peak is not None:
                elapsed = end - start
                elapsed = elapsed.seconds
                with open(self.logs, "a") as logs:
                    logs.write("%s %d" % (elapsed, mem_peak))

        return status == "0"

    def update_job_status(self, status, id_process=None):
        """
        Update job status

        :param status: new status
        :param id_process: system process id
        """
        if MODE == "webserver":
            with Job.connect():
                job = Job.get(Job.id_job == self.id_job)
                job.status = status
                if id_process is not None:
                    job.id_process = id_process
                job.save()
        else:
            self.set_status_standalone(status)

    @staticmethod
    def find_error_in_log(log_file):
        """
        Find error in log (for cluster run)

        :param log_file: log file of the job
        :return: error (empty if no error)
        :rtype: str
        """
        error = ""
        with open(log_file, "r") as log:
            for line in log:
                if line.startswith("###ERR### "):
                    error = line[10:].rstrip()
                    break
        return error

    def launch_to_cluster(self, step, batch_system_type, command, args, log_out, log_err):
        """
        Launch a program to the cluster

        :param step: step (prepare, start)
        :type step: str
        :param batch_system_type: slurm or sge
        :type batch_system_type: str
        :param command: program to launch (without arguments)
        :type command: str
        :param args: arguments to use for the program
        :type args: list
        :param log_out: log file for stdout
        :type log_out: str
        :param log_err: log file for stderr
        :type log_err: str
        :return: True if succeed, else False
        :rtype: bool
        """
        import drmaa
        from dgenies.lib.drmaasession import DrmaaSession
        drmaa_session = DrmaaSession()
        s = drmaa_session.session
        jt = s.createJobTemplate()
        jt.remoteCommand = command
        jt.args = args
        jt.jobName = "_".join([step[:2], self.id_job])
        if log_out == log_err:
            jt.joinFiles = True
            jt.outputPath = ":" + log_out
        else:
            jt.joinFiles = False
            jt.outputPath = ":" + log_out
            jt.errorPath = ":" + log_err

        if step == "start":
            memory = self.config.cluster_memory
            if self.query is None:
                memory = self.config.cluster_memory_ava
                if memory > 32:
                    name, order, contigs, reversed_c, abs_start, c_len = Index.load(self.idx_t, False)
                    if c_len <= 500000000:
                        memory = 32
            if memory > self.tool.max_memory:
                memory = self.tool.max_memory
        else:
            memory = 8000

        native_specs = self.config.drmaa_native_specs
        if batch_system_type == "slurm":
            if native_specs == "###DEFAULT###":
                native_specs = "--mem-per-cpu={0} --mincpus={1} --time={2}"
            if step == "prepare":
                jt.nativeSpecification = native_specs.format(memory, 1, "02:00:00")
            elif step == "start":
                jt.nativeSpecification = native_specs.format(memory // self.tool.threads_cluster * 1000,
                                                             self.tool.threads_cluster, "02:00:00")
        elif batch_system_type == "sge":
            if native_specs == "###DEFAULT###":
                native_specs = "-l mem={0},h_vmem={0} -pe parallel_smp {1}"
            if step == "prepare":
                jt.nativeSpecification = native_specs.format(8000, 1)
            elif step == "start":
                jt.nativeSpecification = native_specs.format(
                    memory // self.tool.threads_cluster * 1000, self.tool.threads_cluster)
        jt.workingDirectory = self.output_dir
        jobid = s.runJob(jt)
        self.id_process = jobid

        self.update_job_status("scheduled-cluster" if step == "start" else "prepare-scheduled", jobid)

        retval = s.wait(jobid, drmaa.Session.TIMEOUT_WAIT_FOREVER)
        if retval.hasExited and (self.check_job_status_slurm() if batch_system_type == "slurm" else
        self.check_job_status_sge()):
            if step == "start":
                status = self.check_job_success()
                if status == "no-match":
                    self._set_analytics_job_status("no-match")
            else:
                status = "prepared"
            # job = Job.get(id_job=self.id_job)
            # job.status = status
            # db.commit()
            self.update_job_status(status)
            s.deleteJobTemplate(jt)
            return status == "succeed" or status == "prepared"
        error = self.find_error_in_log(log_err)
        if step == "prepare":
            error += "<br/>Please check your input file and try again."
        self.set_job_status("fail", error)

        s.deleteJobTemplate(jt)
        return False

    def _launch_drmaa(self, batch_system_type):
        """
        Launch the mapping step to a cluster

        :param batch_system_type: slurm or sge
        :return: True if job succeed, else False
        """

        if self.query is not None:
            args = re.sub("{exe}\s?", "", self.tool.command_line).replace("{query}", self.get_query_split())
        else:
            args = re.sub("{exe}\s?", "", self.tool.all_vs_all)
        out_file = self.logs
        if ">" in args:
            out_file = self.paf_raw
            args = args[:args.index(">")]
        args = args.replace("{target}", self.target.get_path()) \
                   .replace("{threads}", str(self.tool.threads_cluster)) \
                   .replace("{options}", self.options) \
                   .replace("{out}", self.paf_raw)

        args = args.split(" ")

        return self.launch_to_cluster(step="start",
                                      batch_system_type=batch_system_type,
                                      command=self.tool.exec,
                                      args=args,
                                      log_out=out_file,
                                      log_err=self.logs)

    def _getting_local_file(self, fasta, type_f):
        """
        Copy temp file to its final location

        :param fasta: fasta file Object
        :type fasta: Fasta
        :param type_f: query or target
        :type type_f: str
        :return: final full path of the file
        :rtype: str
        """
        finale_path = os.path.join(self.output_dir, type_f + "_" + os.path.basename(fasta.get_path()))
        if fasta.is_example():
            shutil.copy(fasta.get_path(), finale_path)
        else:
            if os.path.exists(fasta.get_path()):
                shutil.move(fasta.get_path(), finale_path)
            else:
                other_file = os.path.join(self.output_dir, ("query" if type_f == "target" else "query") + "_" +
                                          os.path.basename(fasta.get_path()))
                if os.path.exists(other_file):
                    shutil.copy(other_file, finale_path)
                else:
                    raise Exception("Unable to copy %s file from temp to finale path: %s file does not exists" %
                                    (type_f, fasta.get_path()))

        with open(os.path.join(self.output_dir, "." + type_f), "w") as save_file:
            save_file.write(finale_path)
        return finale_path

    def _get_filename_from_url(self, url):
        """
        Retrieve filename from an URL (http or ftp)

        :param url: url of the file to download
        :type url: str
        :return: filename
        :rtype: str
        """
        if url not in self._filename_for_url:
            if url.startswith("ftp://"):
                self._filename_for_url[url] = url.split("/")[-1]
            elif url.startswith("http://") or url.startswith("https://"):
                self._filename_for_url[url] = requests.head(url, allow_redirects=True).url.split("/")[-1]
            else:
                return None
        return self._filename_for_url[url]

    def _download_file(self, url):
        """
        Download a file from an URL

        :param url: url of the file to download
        :type url: str
        :return: absolute path of the downloaded file
        :rtype: str
        """
        local_filename = os.path.join(self.output_dir, self._get_filename_from_url(url))
        # NOTE the stream=True parameter
        if url.startswith("ftp://"):
            urlretrieve(url, local_filename)
        else:
            r = requests.get(url, stream=True)
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        # f.flush() commented by recommendation from J.F.Sebastian
        return local_filename

    def _getting_file_from_url(self, fasta, type_f):
        """
        Download file from URL

        :param fasta: Fasta object describing the input file
        :type fasta: Fasta
        :param type_f: type of the file (query or target)
        :type type_f: str
        :return:
            * [0] True if no error happened, else False
            * [1] If an error happened, True if the error was saved for the job, else False (will be saved later)
            * [2] Finale path of the downloaded file {str}
            * [3] Name of the downloaded file {str}
        :rtype: tuple
        """
        try:
            dl_path = self._download_file(fasta.get_path())
        except (ConnectionError, URLError):
            status = "fail"
            error = "<p>Url <b>%s</b> is not valid!</p>" \
                    "<p>If this is unattended, please contact the support.</p>" % fasta.get_path()
            if MODE == "webserver":
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    job.status = status
                    job.error = error
                    job.save()
            else:
                self.set_status_standalone(status, error)
            return False, True, None, None
        filename = os.path.basename(dl_path)
        name = os.path.splitext(filename.replace(".gz", ""))[0]
        finale_path = os.path.join(self.output_dir, type_f + "_" + filename)
        shutil.move(dl_path, finale_path)
        with open(os.path.join(self.output_dir, "." + type_f), "w") as save_file:
            save_file.write(finale_path)
        return True, False, finale_path, name

    def _check_url(self, fasta, formats):
        """
        Check if an URL is valid, and if the file is valid too

        :param fasta: fasta file object
        :type fasta: Fasta
        :param formats: allowed file formats
        :type formats: tuple
        :return: True if URL and file are valid, else False
        :rtype: bool
        """
        url = fasta.get_path()
        try:
            filename = self._get_filename_from_url(url)
        except (ConnectionError, URLError):
            status = "fail"
            error = "<p>Url <b>%s</b> is not valid!</p>" \
                    "<p>If this is unattended, please contact the support.</p>" % fasta.get_path()
            if MODE == "webserver":
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    job.status = status
                    job.error = error
                    job.save()
            else:
                self.set_status_standalone(status, error)
            return False
        if filename is not None:
            allowed = Functions.allowed_file(filename, formats)
            if not allowed:
                status = "fail"
                format_txt = ""
                if len(formats) == 1:
                    if formats[0] == "fasta":
                        format_txt = "a Fasta file"
                    elif formats[0] == "idx":
                        format_txt = "an index file"
                    elif formats[0] == "map":
                        format_txt = "an alignment file"
                    elif formats[0] == "backup":
                        format_txt = "a backup file"
                    else:
                        format_txt = "a valid file"
                else:
                    if "fasta" in formats and "idx" in formats:
                        format_txt = "a Fasta file or an index file"
                    else:
                        format_txt = "a valid file"
                error = "<p>File <b>%s</b> downloaded from <b>%s</b> is not %s!</p>" \
                        "<p>If this is unattended, please contact the support.</p>" % (filename, url, format_txt)
                if MODE == "webserver":
                    with Job.connect():
                        job = Job.get(Job.id_job == self.id_job)
                        job.status = status
                        job.error = error
                        job.save()
                else:
                    self.set_status_standalone(status, error)
        else:
            allowed = False
            status = "fail"
            error = "<p>Url <b>%s</b> is not a valid URL!</p>" \
                    "<p>If this is unattended, please contact the support.</p>" % url
            if MODE == "webserver":
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    job.status = status
                    job.error = error
                    job.save()
            else:
                self.set_status_standalone(status, error)
        return allowed

    def clear(self):
        """
        Remove job dir
        """
        shutil.rmtree(self.output_dir)

    @staticmethod
    def get_pending_local_number():
        """
        Get number of of jobs running or waiting for a run

        :return: number of jobs
        :rtype: int
        """
        if MODE == "webserver":
            with Job.connect():
                return len(Job.select().where((Job.batch_type == "local") & (Job.status != "success") &
                                              (Job.status != "fail") & (Job.status != "no-match")))
        else:
            return 0

    def set_job_status(self, status, error=""):
        """
        Change status of a job

        :param status: new job status
        :type status: str
        :param error: error description (if any)
        :type error: str
        """
        if MODE == "webserver":
            job = Job.get(Job.id_job == self.id_job)
            job.status = status
            job.error = error
            job.save()
        else:
            self.set_status_standalone(status, error)

    def check_file(self, input_type, should_be_local, max_upload_size_readable):
        """
        Check if file is correct: format, size, valid gzip

        :param input_type: query or target
        :param should_be_local: True if job should be treated locally
        :param max_upload_size_readable: max upload size human readable
        :return: (True if correct, True if error set [for fail], True if should be local)
        """
        if input_type == "target" and self.query is None:
            max_upload_size_readable = self.config.max_upload_size_ava / 1024 / 1024
        with Job.connect():
            my_input = getattr(self, input_type)
            if my_input.get_path().endswith(".gz") and not self.is_gz_file(my_input.get_path()):
                # Check file is correctly gzipped
                self.set_job_status("fail", input_type + " file is not a correct gzip file")
                self.clear()
                return False, True, None
            # Check size:
            file_size = self.get_file_size(my_input.get_path())
            if -1 < (self.config.max_upload_size if (input_type == "query" or self.query is not None)
                     else self.config.max_upload_size_ava) < file_size:
                self.set_job_status("fail",
                                    input_type +
                                    " file exceed size limit of %d Mb (uncompressed)" % max_upload_size_readable)
                self.clear()
                return False, True, None

            if input_type == "align":
                if not hasattr(validators, self.aln_format):
                    self.set_job_status("fail", "Alignment file format not supported")
                    return False, True, None
                if not getattr(validators, self.aln_format)(self.align.get_path()):
                    self.set_job_status("fail", "Alignment file is invalid. Please check your file.")
                    return False, True, None
            elif input_type != "backup":
                if my_input.get_path().endswith(".idx"):
                    if not validators.v_idx(my_input.get_path()):
                        self.set_job_status("fail",
                                            "%s index file is invalid. Please check your file." %
                                            input_type.capitalize())
                        return False, True, None
                if self.config.batch_system_type != "local" and file_size >= getattr(self.config,
                                                                                     "min_%s_size" % input_type):
                    should_be_local = False

        return True, False, should_be_local

    def download_files_with_pending(self, files_to_download, should_be_local, max_upload_size_readable):
        """
        Download files from URLs, with pending (according to the max number of concurrent downloads)

        :param files_to_download: files to download. For each item of the list, it's a list with 2 elements: first one
            is the Fasta object, second one the input type (query or target)
        :type files_to_download: list of list
        :param should_be_local: True if the job should be run locally (according to input file sizes), else False
        :type should_be_local: bool
        :param max_upload_size_readable: Human readable max upload size (to show on errors)
        :type max_upload_size_readable: str
        """
        with Job.connect():
            status = "getfiles-waiting"
            if MODE == "webserver":
                job = Job.get(Job.id_job == self.id_job)
                job.status = status
                job.save()
                # Create a session:
                s_id = Session.new(True)
                session = Session.get(s_id=s_id)
            else:
                status = "getfiles"
                session = None
                job = None
                s_id = None

            try:
                correct = True
                error_set = False
                if MODE == "webserver":
                    allowed = session.ask_for_upload(True)
                else:
                    allowed = True
                while not allowed:
                    time.sleep(15)
                    session = Session.get(s_id=s_id)
                    allowed = session.ask_for_upload(False)
                if allowed:
                    if MODE == "webserver":
                        job.status = "getfiles"
                        job.save()
                    for file, input_type in files_to_download:
                        correct, error_set, finale_path, filename = self._getting_file_from_url(file, input_type)
                        if not correct:
                            break
                        my_input = getattr(self, input_type)
                        my_input.set_path(finale_path)
                        my_input.set_name(filename)
                        correct, error_set, should_be_local = self.check_file(input_type, should_be_local,
                                                                              max_upload_size_readable)
                        if not correct:
                            break

                    if correct and MODE == "webserver" and job.batch_type != "local" and should_be_local \
                            and self.get_pending_local_number() < self.config.max_run_local:
                        job.batch_type = "local"
                        job.save()
                else:
                    correct = False
            except:  # Except all possible exceptions
                traceback.print_exc()
                correct = False
                error_set = False
            if MODE == "webserver":
                session.delete_instance()
            self._after_start(correct, error_set)

    def getting_files(self):
        """
        Get files for the job

        :return:
            * [0] True if getting files succeed, False else
            * [1] If error happenned, True if error already saved for the job, False else (error will be saved later)
            * [2] True if no data must be downloaded (will be downloaded with pending if True)
        :rtype: tuple
        """
        with Job.connect():
            status = "getfiles"
            if MODE == "webserver":
                job = Job.get(Job.id_job == self.id_job)
                job.status = status
                job.save()
            else:
                job = None
                self.set_status_standalone(status)
            correct = True
            error_set = False
            should_be_local = True
            max_upload_size_readable = self.config.max_upload_size / 1024 / 1024  # Set it in Mb
            files_to_download = []
            if self.query is not None:
                if self.query.get_type() == "local":
                    self.query.set_path(self._getting_local_file(self.query, "query"))
                    correct, error_set, should_be_local = self.check_file("query", should_be_local,
                                                                          max_upload_size_readable)
                    if not correct:
                        return False, error_set, True
                elif self._check_url(self.query, ("fasta",) if self.align is None else ("fasta", "idx")):
                    files_to_download.append([self.query, "query"])
                else:
                    return False, True, True
            if correct and self.target is not None:
                if self.target.get_type() == "local":
                    self.target.set_path(self._getting_local_file(self.target, "target"))
                    correct, error_set, should_be_local = self.check_file("target", should_be_local,
                                                                          max_upload_size_readable)
                    if not correct:
                        return False, error_set, True
                elif self._check_url(self.target, ("fasta",) if self.align is None else ("fasta", "idx")):
                    files_to_download.append([self.target, "target"])
                else:
                    return False, True, True
            if correct and self.align is not None:
                if self.align.get_type() == "local":
                    self.align.set_path(self._getting_local_file(self.align, "align"))
                    correct, error_set, should_be_local = self.check_file("align", should_be_local,
                                                                          max_upload_size_readable)
                elif self._check_url(self.align, ("map",)):
                    files_to_download.append([self.align, "align"])
                else:
                    return False, True, True
            if correct and self.backup is not None:
                if self.backup.get_type() == "local":
                    self.backup.set_path(self._getting_local_file(self.backup, "backup"))
                    correct, error_set, should_be_local = self.check_file("backup", should_be_local,
                                                                          max_upload_size_readable)
                elif self._check_url(self.backup, ("backup",)):
                    files_to_download.append([self.backup, "backup"])
                else:
                    return False, True, True

            all_downloaded = True
            if correct:
                if len(files_to_download) > 0:
                    all_downloaded = False
                    thread = threading.Timer(0, self.download_files_with_pending,
                                             kwargs={"files_to_download": files_to_download,
                                                     "should_be_local": should_be_local,
                                                     "max_upload_size_readable": max_upload_size_readable})
                    thread.start()  # Start the execution

                elif correct and MODE == "webserver" and job.batch_type != "local" and should_be_local \
                        and self.get_pending_local_number() < self.config.max_run_local:
                    job.batch_type = "local"
                    job.save()
        return correct, error_set, all_downloaded

    def send_mail_post(self):
        """
        Send mail using POST url (if there is no access to mailer)
        """
        key = Functions.random_string(15)
        key_file = os.path.join(self.config.app_data, self.id_job, ".key")
        with open(key_file, "w") as k_f:
            k_f.write(key)
        data = parse.urlencode({"key": key}).encode()
        req = request.Request(self.config.web_url + "/send-mail/" + self.id_job, data=data)
        resp = request.urlopen(req)
        if resp.getcode() != 200:
            print("Job %s: Send mail failed!" % self.id_job)

    def run_job_in_thread(self, batch_system_type="local"):
        """
        Run a job asynchronously into a new thread

        :param batch_system_type: slurm or sge
        :type batch_system_type: str
        """
        thread = threading.Timer(1, self.run_job, kwargs={"batch_system_type": batch_system_type})
        thread.start()  # Start the execution

    def prepare_data_in_thread(self):
        """
        Prepare data in a new thread
        """
        thread = threading.Timer(1, self.prepare_data)
        thread.start()  # Start the execution

    def prepare_data_cluster(self, batch_system_type):
        """
        Launch of prepare data on a cluster

        :param batch_system_type: slurm or sge
        :type batch_system_type: str
        :return: True if succeed, else False
        :rtype: bool
        """
        args = [self.config.cluster_prepare_script,
                "-t", self.target.get_path(),
                "-m", self.target.get_name(),
                "-p", self.preptime_file]
        if self.query is not None:
            args += ["-q", self.query.get_path(),
                     "-u", self.get_query_split(),
                     "-n", self.query.get_name()]
            if self.tool.split_before:
                args.append("--split")
        success = self.launch_to_cluster(step="prepare",
                                         batch_system_type=batch_system_type,
                                         command=self.config.cluster_python_exec,
                                         args=args,
                                         log_out=self.logs,
                                         log_err=self.logs)
        if not success and MODE == "webserver" and self.config.send_mail_status:
            self.send_mail_post()
        return success

    def prepare_data_local(self):
        """
        Prepare data locally. On standalone mode, launch job after, if success.
        :return: True if job succeed, else False
        :rtype: bool
        """
        with open(self.preptime_file, "w") as ptime, Job.connect():
            self.set_job_status("preparing")
            ptime.write(str(round(time.time())) + "\n")
            error_tail = "Please check your input file and try again."
            if self.query is not None:
                fasta_in = self.query.get_path()
                if self.tool.split_before:
                    split = True
                    splitter = Splitter(input_f=fasta_in, name_f=self.query.get_name(), output_f=self.get_query_split(),
                                        query_index=self.query_index_split, debug=DEBUG)
                    success, error = splitter.split()
                    nb_contigs = splitter.nb_contigs
                    in_fasta = self.get_query_split()
                else:
                    split = False
                    uncompressed = None
                    if self.query.get_path().endswith(".gz"):
                        uncompressed = self.query.get_path()[:-3]
                    success, nb_contigs, error = index_file(self.query.get_path(), self.query.get_name(), self.idx_q,
                                                            uncompressed)
                    in_fasta = self.query.get_path()
                    if uncompressed is not None:
                        in_fasta = uncompressed
                if success:
                    filtered_fasta = os.path.join(os.path.dirname(self.get_query_split()), "filtered_" +
                                                  os.path.basename(self.get_query_split()))
                    filter_f = Filter(fasta=in_fasta,
                                      index_file=self.query_index_split if split else self.idx_q,
                                      type_f="query",
                                      min_filtered=round(nb_contigs / 4),
                                      split=True,
                                      out_fasta=filtered_fasta,
                                      replace_fa=True)
                    filter_f.filter()
                else:
                    self.set_job_status("fail", "<br/>".join(["Query fasta file is not valid:", error, error_tail]))
                    if self.config.send_mail_status:
                        self.send_mail_post()
                    return False
            uncompressed = None
            if self.target.get_path().endswith(".gz"):
                uncompressed = self.target.get_path()[:-3]
            success, nb_contigs, error = index_file(self.target.get_path(), self.target.get_name(), self.idx_t, uncompressed)
            if success:
                in_fasta = self.target.get_path()
                if uncompressed is not None:
                    in_fasta = uncompressed
                filtered_fasta = os.path.join(os.path.dirname(in_fasta), "filtered_" + os.path.basename(in_fasta))
                filter_f = Filter(fasta=in_fasta,
                                  index_file=self.idx_t,
                                  type_f="target",
                                  min_filtered=round(nb_contigs / 4),
                                  split=False,
                                  out_fasta=filtered_fasta,
                                  replace_fa=True)
                is_filtered = filter_f.filter()
                if uncompressed is not None:
                    if is_filtered:
                        os.remove(self.target.get_path())
                        self.target.set_path(uncompressed)
                        with open(os.path.join(self.output_dir, ".target"), "w") as save_file:
                            save_file.write(uncompressed)
                    else:
                        os.remove(uncompressed)
            else:
                if uncompressed is not None:
                    try:
                        os.remove(uncompressed)
                    except FileNotFoundError:
                        pass
                self.set_job_status("fail", "<br/>".join(["Target fasta file is not valid:", error, error_tail]))
                if self.config.send_mail_status:
                    self.send_mail_post()
                return False
            ptime.write(str(round(time.time())) + "\n")
            self.set_job_status("prepared")
            if MODE != "webserver":
                self.run_job("local")
            return True

    def _end_of_prepare_dotplot(self):
        """
        Tasks done after preparing dot plot data: parse & sort of alignment file
        """
        # Parse alignment file:
        if hasattr(parsers, self.aln_format):
            getattr(parsers, self.aln_format)(self.align.get_path(), self.paf_raw)
            os.remove(self.align.get_path())
        elif self.aln_format == "paf":
            shutil.move(self.align.get_path(), self.paf_raw)
        else:
            self.set_job_status("fail", "No parser found for format %s. Please contact the support." % self.aln_format)
            return False

        self.set_job_status("started")

        # Sort paf lines:
        sorter = Sorter(self.paf_raw, self.paf)
        sorter.sort()
        os.remove(self.paf_raw)
        if self.target is not None and os.path.exists(self.target.get_path()) and not \
                self.target.get_path().endswith(".idx"):
            os.remove(self.target.get_path())

        self.align.set_path(self.paf)

        self.set_job_status("success")

        if MODE == "webserver" and self.config.send_mail_status:
            self.send_mail_post()

        return True

    def prepare_dotplot_cluster(self, batch_system_type):
        """
        Prepare data if alignment already done: just index the fasta (if index not given), then parse the alignment

        :param batch_system_type: type of cluster (slurm or sge)
        :type batch_system_type: str
        """

        args = [self.config.cluster_prepare_script,
                "-p", self.preptime_file, "--index-only"]

        has_index = False

        target_format = os.path.splitext(self.target.get_path())[1][1:]
        if target_format == "idx":
            shutil.move(self.target.get_path(), self.idx_t)
            os.remove(os.path.join(self.output_dir, ".target"))
        else:
            has_index = True
            args += ["-t", self.target.get_path(),
                     "-m", self.target.get_name()]
        if self.query is not None:
            query_format = os.path.splitext(self.query.get_path())[1][1:]
            if query_format == "idx":
                shutil.move(self.query.get_path(), self.idx_q)
                os.remove(os.path.join(self.output_dir, ".query"))
            else:
                has_index = True
                args += ["-q", self.query.get_path(),
                         "-n", self.query.get_name()]

        success = True
        if has_index:
            success = self.launch_to_cluster(step="prepare",
                                             batch_system_type=batch_system_type,
                                             command=self.config.cluster_python_exec,
                                             args=args,
                                             log_out=self.logs,
                                             log_err=self.logs)

        if success:
            if self.query is None:
                shutil.copy(self.idx_t, self.idx_q)
            return self._end_of_prepare_dotplot()
        elif MODE == "webserver" and self.config.send_mail_status:
            self.send_mail_post()

        return False

    def prepare_dotplot_local(self):
        """
        Prepare data if alignment already done: just index the fasta (if index not given), then parse the alignment
        file and sort it.
        """
        self.set_job_status("preparing")
        # Prepare target index:
        target_format = os.path.splitext(self.target.get_path())[1][1:]
        if target_format == "idx":
            shutil.move(self.target.get_path(), self.idx_t)
            os.remove(os.path.join(self.output_dir, ".target"))
        else:
            index_file(self.target.get_path(), self.target.get_name(), self.idx_t)

        # Prepare query index:
        if self.query is not None:
            query_format = os.path.splitext(self.query.get_path())[1][1:]
            if query_format == "idx":
                shutil.move(self.query.get_path(), self.idx_q)
                os.remove(os.path.join(self.output_dir, ".query"))
            else:
                index_file(self.query.get_path(), self.query.get_name(), self.idx_q)
        else:
            shutil.copy(self.idx_t, self.idx_q)

        return self._end_of_prepare_dotplot()

    def prepare_data(self):
        """
        Launch preparation of data
        """
        if self.align is None:
            if MODE == "webserver":
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    if job.batch_type == "local":
                        success = self.prepare_data_local()
                    else:
                        success = self.prepare_data_cluster(job.batch_type)
                    if not success:
                        self._set_analytics_job_status("fail-prepare")
            else:
                self.prepare_data_local()
        else:
            if MODE == "webserver":
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    if job.batch_type == "local":
                        success = self.prepare_dotplot_local()
                    else:
                        success = self.prepare_dotplot_cluster(job.batch_type)
                    if success:
                        self._set_analytics_job_status("success")
                    else:
                        self._set_analytics_job_status("fail-all")
            else:
                self.prepare_dotplot_local()

    def run_job(self, batch_system_type):
        """
        Run of a job (mapping step)

        :param batch_system_type: type of cluster (slurm or sge)
        :type batch_system_type: str
        """
        try:
            success = False
            if batch_system_type == "local":
                success = self._launch_local()
            elif batch_system_type in ["slurm", "sge"]:
                success = self._launch_drmaa(batch_system_type)
            if success:
                with Job.connect():
                    if MODE == "webserver":
                        job = Job.get(Job.id_job == self.id_job)
                        with open(self.logs) as logs:
                            measures = logs.readlines()[-1].strip("\n").split(" ")
                            map_elapsed = round(float(measures[0]))
                            job.mem_peak = int(measures[1])
                        with open(self.preptime_file) as ptime:
                            lines = ptime.readlines()
                            start = int(lines[0].strip("\n"))
                            end = int(lines[1].strip("\n"))
                            prep_elapsed = end - start
                            job.time_elapsed = prep_elapsed + map_elapsed
                    else:
                        job = None
                    status = "merging"
                    if MODE == "webserver":
                        job.status = "merging"
                        job.save()
                    else:
                        self.set_status_standalone(status)
                    if self.tool.split_before and self.query is not None:
                        start = time.time()
                        paf_raw = self.paf_raw + ".split"
                        os.remove(self.get_query_split())
                        merger = Merger(self.paf_raw, paf_raw, self.query_index_split,
                                        self.idx_q, debug=DEBUG)
                        merger.merge()
                        os.remove(self.paf_raw)
                        os.remove(self.query_index_split)
                        self.paf_raw = paf_raw
                        end = time.time()
                        if MODE == "webserver":
                            job.time_elapsed += end - start
                    elif self.query is None:
                        shutil.copyfile(self.idx_t, self.idx_q)
                        Path(os.path.join(self.output_dir, ".all-vs-all")).touch()
                    if self.tool.parser is not None:
                        paf_raw = self.paf_raw + ".parsed"
                        getattr(parsers, self.tool.parser)(self.paf_raw, paf_raw)
                        os.remove(self.paf_raw)
                        self.paf_raw = paf_raw
                    sorter = Sorter(self.paf_raw, self.paf)
                    sorter.sort()
                    os.remove(self.paf_raw)
                    if self.target is not None and os.path.exists(self.target.get_path()):
                        os.remove(self.target.get_path())
                    if os.path.isfile(os.path.join(self.output_dir, ".do-sort")):
                        paf = Paf(paf=self.paf,
                                  idx_q=self.idx_q,
                                  idx_t=self.idx_t,
                                  auto_parse=False)
                        paf.sort()
                        if not paf.parsed:
                            success = False
                            status = "fail"
                            error = "Error while sorting query. Please contact us to report the bug"
                            if MODE == "webserver":
                                job = Job.get(Job.id_job == self.id_job)
                                job.status = status
                                job.error = error
                                self._set_analytics_job_status("fail-sort")
                            else:
                                self.set_status_standalone(status, error)
                    if success:
                        status = "success"
                        if MODE == "webserver":
                            job = Job.get(Job.id_job == self.id_job)
                            job.status = "success"
                            job.save()

                            # Analytics:
                            self._set_analytics_job_status("success")
                            if self.config.send_mail_status:
                                self.send_mail_post()

                        else:
                            self.set_status_standalone(status)
            elif MODE == "webserver":
                self._set_analytics_job_status("fail-map")
        except Exception as e:
            traceback.print_exc()
            with open(self.logs, 'a') as f:
                f.write(str(e))
                f.write(traceback.format_exc())
            self.set_job_status("fail", "Your job has failed for an unexpected reason. Please contact the support if"
                                        "the problem persists.")
            if MODE == "webserver":
                self._set_analytics_job_status("fail-map-after")

    def _anonymize_mail_client(self, email):
        """
        Replace the email address with its group defined in config file if anonymization is enabled
        :param email: email to anonymize
        :type email: str
        :return: email group if anonymization is enabled (empty string if no group matching), email else
        :rtype: str
        """
        if not self.config.anonymous_analytics:
            return email
        for group, pattern in self.config.analytics_groups:
            if re.match(pattern, email):
                return group
        return ''

    def _save_analytics_data(self):
        """
        Save analytics data into the database
        """
        from dgenies.database import Analytics
        with Job.connect():
            job = Job.get(Job.id_job == self.id_job)
            target_size = os.path.getsize(self.target.get_path()) if (self.target is not None and self.target.get_type()
                                                                      == "local" and
                                                                      os.path.exists(self.target.get_path())) else 0
            query_size = None
            if self.query is not None and self.query.get_type() == "local" and os.path.exists(self.query.get_path()):
                query_size = os.path.getsize(self.query.get_path())
            log = Analytics.create(
                id_job=self.id_job,
                date_created=datetime.now(),
                target_size=target_size,
                query_size=query_size,
                mail_client=self._anonymize_mail_client(job.email),
                batch_type=job.batch_type,
                job_type="new" if (self.align is None and self.backup is None) else "plot",
                tool=self.tool_name if self.tool_name is not None else "unset")
            log.save()

    def _set_analytics_job_status(self, status):
        """
        Change status for a job in analytics database

        :param status: new status
        :type status: str (20)
        """
        if self.config.analytics_enabled and MODE == "webserver":
            from dgenies.database import Analytics
            with Job.connect():
                analytic = Analytics.get(Analytics.id_job == self.id_job)
                if analytic.status != "no-match":
                    analytic.status = status
                    analytic.save()

    def unpack_backup(self):
        """
        Untar backup file
        """
        try:
            with tarfile.open(self.backup.get_path(), "r:*") as tar:
                names = tar.getnames()
                if len(names) != 3:
                    return False
                for name in ["map.paf", "query.idx", "target.idx"]:
                    if name not in names:
                        return False
                tar.extractall(self.output_dir)
                shutil.move(self.paf, self.paf_raw)
                if not validators.paf(self.paf_raw):
                    return False
                target_path = os.path.join(self.output_dir, "target.idx")
                query_path = os.path.join(self.output_dir, "query.idx")
                if not validators.v_idx(target_path) or not validators.v_idx(query_path):
                    return False
                self.align = Fasta(name="map", path=self.paf_raw, type_f="local")
                self.aln_format = "paf"
                with open(os.path.join(self.output_dir, ".align"), "w") as aln:
                    aln.write(self.paf_raw)
                self.target = Fasta(name="target", path=target_path, type_f="local")
                with open(os.path.join(self.output_dir, ".target"), "w") as trgt:
                    trgt.write(target_path)
                self.query = Fasta(name="query", path=query_path, type_f="local")
                with open(os.path.join(self.output_dir, ".query"), "w") as qr:
                    qr.write(query_path)
            os.remove(self.backup.get_path())
            return True
        except:
            traceback.print_exc()
            return False

    def _after_start(self, success, error_set):
        """
        Tasks done after input files downloaded, checked and parsed: if success, set job status to "waiting".
        Else, set job error and send mail.

        :param success: job success
        :type success: bool
        :param error_set: error already set (else, set it now)
        :type error_set: bool
        """
        with Job.connect():
            if MODE == "webserver" and self.config.analytics_enabled:
                self._save_analytics_data()
            if success:
                if self.backup is not None:
                    success = self.unpack_backup()
                    if not success:
                        self.set_job_status("fail", "Backup file is not valid. If it is unattended, please contact the "
                                                    "support.")
                        if MODE == "webserver" and self.config.send_mail_status:
                            self.send_mail()
                        return False
                status = "waiting"
                if MODE == "webserver":
                    job = Job.get(Job.id_job == self.id_job)
                    job.status = status
                    job.save()
                else:
                    self.set_status_standalone(status)
                    self.prepare_data_in_thread()
            else:
                self._set_analytics_job_status("fail-getfiles")
                if not error_set:
                    status = "fail"
                    error = "<p>Error while getting input files. Please contact the support to report the bug.</p>"
                    if MODE == "webserver":
                        job = Job.get(Job.id_job == self.id_job)
                        job.status = status
                        job.error = error
                        job.save()
                    else:
                        self.set_status_standalone(status, error)
                if MODE == "webserver" and self.config.send_mail_status:
                    self.send_mail()

    def start_job(self):
        """
        Start job: download, check and parse input files
        """
        try:
            success, error_set, all_downloaded = self.getting_files()
            if not success or all_downloaded:
                self._after_start(success, error_set)

        except Exception:
            traceback.print_exc()
            error = "<p>An unexpected error has occurred. Please contact the support to report the bug.</p>"
            if MODE == "webserver":
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    job.status = "fail"
                    job.error = error
                    job.save()
                    if self.config.send_mail_status:
                        self.send_mail()
            else:
                self.set_status_standalone("fail", error)

    def launch_standalone(self):
        """
        Launch a job in standalone mode (asynchronously in a new thread)
        """
        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)
        self.set_status_standalone("submitted")
        thread = threading.Timer(1, self.start_job)
        thread.start()

    def launch(self):
        """
        Launch a job in webserver mode (asynchronously in a new thread)
        """
        with Job.connect():
            j1 = Job.select().where(Job.id_job == self.id_job)
            if len(j1) > 0:
                print("Old job found without result dir existing: delete it from BDD!")
                for j11 in j1:
                    j11.delete_instance()
            if self.target is not None or self.backup is not None:
                job = Job.create(id_job=self.id_job, email=self.email, batch_type=self.config.batch_system_type,
                                 date_created=datetime.now(), tool=self.tool.name if self.tool is not None else None,
                                 options=self.options)
                job.save()
                if not os.path.exists(self.output_dir):
                    os.mkdir(self.output_dir)
                thread = threading.Timer(1, self.start_job)
                thread.start()
            else:
                job = Job.create(id_job=self.id_job, email=self.email, batch_type=self.config.batch_system_type,
                                 date_created=datetime.now(), tool=self.tool.name if self.tool is not None else None,
                                 options=self.options, status="fail")
                job.save()

    def set_status_standalone(self, status, error=""):
        """
        Change job status in standalone mode

        :param status: new status
        :type status: str
        :param error: error description (if any)
        :type error: str
        """
        status_file = os.path.join(self.output_dir, ".status")
        with open(status_file, "w") as s_file:
            s_file.write("|".join([status, error]))

    def get_status_standalone(self, with_error=False):
        """
        Get job status in standalone mode

        :param with_error: get also the error
        :return: status (and error, if with_error=True)
        :rtype: str or tuple (if with_error=True)
        """
        status_file = os.path.join(self.output_dir, ".status")
        with open(status_file, "r") as s_file:
            items = s_file.read().strip("\n").split("|")
            if with_error:
                return items
            return items[0]

    def status(self):
        """
        Get job status and error. In webserver mode, get also mem peak and time elapsed

        :return: status and other information
        :rtype: dict
        """
        if MODE == "webserver":
            try:
                with Job.connect():
                    job = Job.get(Job.id_job == self.id_job)
                    return {"status": job.status, "mem_peak": job.mem_peak, "time_elapsed": job.time_elapsed,
                            "error": job.error}
            except DoesNotExist:
                return {"status": "unknown", "error": ""}
        else:
            try:
                status, error = self.get_status_standalone(True)
                return {"status": status, "mem_peak": None, "time_elapsed": None, "error": error}
            except FileNotFoundError:
                return {"status": "unknown", "error": ""}

    def delete(self):
        """
        Remove a job

        :return:
            * [0] Success of the deletion
            * [1] Error message, if any (else empty string)
        :rtype: (bool, str)
        """
        if not os.path.exists(self.output_dir) or not os.path.isdir(self.output_dir):
            return False, "Job does not exists"
        if MODE == "webserver":
            try:
                job = Job.get(id_job=self.id_job)
            except DoesNotExist:
                pass
            else:
                is_gallery = Gallery.select().where(Gallery.job == job)
                if is_gallery:
                    return False, "Delete a job that is in gallery is forbidden"
                job.delete_instance()
        shutil.rmtree(self.output_dir)
        return True, ""



#!/usr/bin/env python3

import os
import time
import psutil
import atexit
from datetime import datetime
from tendo import singleton
import argparse

from dgenies.database import Job, Session
from dgenies.config_reader import AppConfigReader
from dgenies.lib.job_manager import JobManager

# Allow only one instance:
me = singleton.SingleInstance()

config_reader = AppConfigReader()

# Add DRMAA lib in env:
if config_reader.drmaa_lib_path is not None and config_reader.batch_system_type != "local":
    os.environ["DRMAA_LIBRARY_PATH"] = config_reader.drmaa_lib_path
    try:
        import drmaa
        from dgenies.lib.drmaasession import DrmaaSession
        DRMAA_SESSION = DrmaaSession()
    except ImportError:
        pass

NB_RUN = config_reader.local_nb_runs
NB_PREPARE = config_reader.nb_data_prepare

DEBUG = config_reader.debug

LOG_FILE = "stdout"


def _printer(*messages):
    """
    print messages to stdout or to a file (according to LOG_FILE global constant)

    :param messages: messages to print
    """
    if DEBUG:
        if LOG_FILE == "stdout":
            print(*messages)
        else:
            with open(LOG_FILE, "a") as log_f:
                print(*messages, file=log_f)


def start_job(id_job, batch_system_type="local"):
    """
    Start a job (mapping step)

    :param id_job: job id
    :type id_job: str
    :param batch_system_type: local, slurm or sge
    :type batch_system_type: str
    """
    _printer("Start job", id_job)
    with Job.connect():
        job = Job.get(Job.id_job == id_job)
        job.status = "starting"
        job.save()
        job_mng = JobManager(id_job=id_job, email=job.email, tool=job.tool, options=job.options)
        job_mng.set_inputs_from_res_dir()
        job_mng.run_job_in_thread(batch_system_type)


def get_scheduled_local_jobs():
    """
    Get list of jobs ready to be started (for local runs)

    :return: list of jobs
    :rtype: list
    """
    all_jobs = []
    with Job.connect():
        jobs = Job.select().where((Job.batch_type == "local") & ((Job.status == "prepared") | (Job.status == "scheduled"))).\
            order_by(Job.date_created)
        for job in jobs:
            all_jobs.append(job.id_job)
            job.status = "scheduled"
            job.save()
    return all_jobs


def get_scheduled_cluster_jobs():
    """
    Get list of jobs ready to be started (for cluster runs)

    :return: list of jobs
    :rtype: list
    """
    all_jobs = []
    with Job.connect():
        jobs = Job.select().where((Job.batch_type != "local") & ((Job.status == "prepared") | (Job.status == "scheduled"))).\
            order_by(Job.date_created)
        for job in jobs:
            all_jobs.append({"job_id": job.id_job, "batch_type": job.batch_type})
            job.status = "scheduled"
            job.save()
    return all_jobs


def prepare_job(id_job):
    """
    Launch job preparation of data

    :param id_job: job id
    :type id_job: str
    """
    _printer("Prepare data for job:", id_job)
    with Job.connect():
        job = Job.get(Job.id_job == id_job)
        job.status = "preparing"
        job.save()
        job_mng = JobManager(id_job=id_job, email=job.email, tool=job.tool)
#        job_mng = JobManager(id_job=id_job, email=job.email, tool=job.tool, options=job.options)
        job_mng.set_inputs_from_res_dir()
        job_mng.prepare_data_in_thread()


def get_prep_scheduled_jobs():
    """
    Get list of jobs ready to be prepared (all data is downloaded and parsed)

    :return: list of jobs
    :rtype: list
    """
    with Job.connect():
        jobs = Job.select().where(Job.status == "waiting").order_by(Job.date_created)
        return [(j.id_job, j.batch_type) for j in jobs]


def get_preparing_jobs_nb():
    """
    Get number of jobs in preparation step (for local runs)

    :return: number of jobs
    :rtype: int
    """
    with Job.connect():
        return len(Job.select().where(Job.status == "preparing"))


def get_preparing_jobs_cluster_nb():
    """
    Get number of jobs in preparation step (for cluster runs)

    :return: number of jobs
    :rtype: int
    """
    with Job.connect():
        return len(Job.select().where(Job.status == "preparing-cluster")), \
               len(Job.select().where(Job.status == "prepare-scheduled"))


def parse_started_jobs():
    """
    Parse all started jobs: check all is OK, change jobs status if needed. Look for died jobs

    :return: (list of id of jobs started locally, list of id of jobs started on cluster)
    :rtype: (list, list)
    """
    with Job.connect():
        jobs_started = []  # Only local jobs
        cluster_jobs_started = []  # Only cluster jobs
        jobs = Job.select().where((Job.status == "started") | (Job.status == "starting") | (Job.status == "succeed") |
                                  (Job.status == "merging") | (Job.status == "scheduled-cluster") |
                                  (Job.status == "prepare-scheduled") | (Job.status == "prepare-cluster"))
        for job in jobs:
            pid = job.id_process
            if job.batch_type == "local":
                if job.status != "started" or psutil.pid_exists(pid):
                    jobs_started.append(job.id_job)
                else:
                    print("Job %s (pid: %d) has died!" % (job.id_job, job.id_process))
                    job.status = "fail"
                    job.error = "<p>Your job has failed for an unexpected reason. Please contact the support.</p>"
                    job.save()
                    # Todo: send mail about the error
            else:
                if job.status in ["started", "scheduled-cluster", "prepare-scheduled", "preparing-cluster"]:
                    s = DRMAA_SESSION.session
                    status = s.jobStatus(str(job.id_process))
                    if status not in [drmaa.JobState.RUNNING, drmaa.JobState.DONE, drmaa.JobState.QUEUED_ACTIVE,
                                      drmaa.JobState.SYSTEM_ON_HOLD, drmaa.JobState.USER_ON_HOLD,
                                      drmaa.JobState.USER_SYSTEM_ON_HOLD]:
                        if job.batch_type == "slurm":
                            os.system("scancel %s" % job.id_process)
                        elif job.batch_type == "sge":
                            os.system("qdel %s" % job.id_process)
                        print("Job %s (id on cluster: %d) has died!" % (job.id_job, job.id_process))
                        job.status = "fail"
                        job.error = "<p>Your job has failed for an unexpected reason. Please contact the support.</p>"
                        job.save()
                        # Todo: send mail about the error
                    else:
                        if job.status == "scheduled-cluster" and status == drmaa.JobState.RUNNING:
                            job.status = "started"
                            job.save()
                            cluster_jobs_started.append(job.id_job)
                        elif job.status == "prepare-scheduled" and status == drmaa.JobState.RUNNING:
                            job.status = "preparing-cluster"
                            job.save()
                        elif job.status == "started":
                            cluster_jobs_started.append(job.id_job)
                else:
                    cluster_jobs_started.append(job.id_job)
    return jobs_started, cluster_jobs_started


def parse_uploads_asks():
    """
    Parse asks for an upload: allow new uploads when other end, remove expired sessions, ...
    """
    with Session.connect():
        now = datetime.now()
        # Get allowed:
        all_sessions = Session.select()
        nb_sessions = len(all_sessions)
        _printer("All sessions:", nb_sessions)
        sessions = Session.select().where(Session.status == "active")
        nb_active_dl = len(sessions)
        _printer("Active_dl:", nb_active_dl)
        for session in sessions:
            if not session.keep_active and (now - session.last_ping).total_seconds() > 50:
                _printer("Delete 1 active session:", session.s_id)
                session.delete_instance()  # We consider the user has left
                nb_active_dl -= 1
        # Get pending:
        sessions = Session.select().where(Session.status == "pending").order_by(Session.date_created)
        _printer("Pending:", len(sessions))
        for session in sessions:
            delay = (now - session.last_ping).total_seconds()
            if delay > 30:
                session.status = "reset"  # Reset position, the user has probably left
                session.save()
                _printer("Reset 1 session:", session.s_id)
            elif nb_active_dl < config_reader.max_concurrent_dl:
                session.status = "active"
                session.save()
                nb_active_dl += 1
                _printer("Enable 1 session:", session.s_id)
        # Remove old sessions:
        for session in all_sessions:
            delay = (now - session.last_ping).total_seconds()
            if delay > 86400:  # Session has more than 1 day
                _printer("Delete 1 outdated session:", session.s_id)
                session.delete_instance()  # Session has expired


@atexit.register
def cleaner():
    """
    Exit DRMAA session at program exit
    """
    if "DRMAA_SESSION" in globals():
        DRMAA_SESSION.exit()


def move_job_to_cluster(id_job):
    """
    Change local job to be run on the cluster

    :param id_job:
    :return:
    """
    with Job.connect():
        job = Job.get(Job.id_job == id_job)
        job.batch_type = config_reader.batch_system_type
        job.save()


def parse_args():
    """
    Parse command line arguments and define DEBUG and LOG_FILE constants
    """

    global DEBUG, LOG_FILE

    parser = argparse.ArgumentParser(description="Start local scheduler")
    parser.add_argument('-d', '--debug', type=str, required=False, help="Set to True to enable debug")
    parser.add_argument('-l', '--log-dir', type=str, required=False, help="Folder into store logs")
    args = parser.parse_args()

    if args.debug is not None:
        if args.debug.lower() == "true" or args.debug.lower == "1":
            DEBUG = True
        elif args.debug.lower() == "false" or args.debug.lower == "0":
            DEBUG = False
        else:
            raise Exception("Invalid value for debug: %s (valid values: True, False)" % args.debug)

    if args.log_dir is not None:
        log_dir = args.log_dir
    else:
        log_dir = config_reader.log_dir

    if DEBUG:
        if log_dir == "stdout":
            LOG_FILE = "stdout"
        else:
            LOG_FILE = os.path.join(config_reader.log_dir, "local_scheduler.log")


if __name__ == '__main__':
    parse_args()

    while True:
        _printer("Check uploads...")
        parse_uploads_asks()
        _printer("")
        _printer("Checking jobs...")
        scheduled_jobs_local = get_scheduled_local_jobs()
        scheduled_jobs_cluster = get_scheduled_cluster_jobs()
        prep_scheduled_jobs = get_prep_scheduled_jobs()
        _printer("Waiting for preparing:", len(prep_scheduled_jobs))
        nb_preparing_jobs = get_preparing_jobs_nb()
        nb_preparing_jobs_cluster = get_preparing_jobs_cluster_nb()
        _printer("Preparing:", nb_preparing_jobs, "(local)", "".join([str(nb_preparing_jobs_cluster[0]),
                 "[", str(nb_preparing_jobs_cluster[1]), "]"]), "(cluster)")
        _printer("Scheduled:", len(scheduled_jobs_local), "(local),", len(scheduled_jobs_cluster), "(cluster)")
        started_jobs, cluster_started_jobs = parse_started_jobs()
        nb_started = len(started_jobs)
        _printer("Started:", nb_started, "(local),", len(cluster_started_jobs), "(cluster)")
        nj = 0
        local_waiting_jobs = []
        while nj < len(prep_scheduled_jobs):
            job_batch_type = prep_scheduled_jobs[nj][1]
            if nb_preparing_jobs < NB_PREPARE or job_batch_type != "local":
                prepare_job(prep_scheduled_jobs[nj][0])
                if job_batch_type == "local":
                    nb_preparing_jobs += 1
                del prep_scheduled_jobs[nj]
            else:
                if job_batch_type == "local":
                    local_waiting_jobs.append(prep_scheduled_jobs[nj][0])
                nj += 1
        if config_reader.batch_system_type != "local" and len(local_waiting_jobs) > config_reader.max_wait_local:
            for id_job in local_waiting_jobs[config_reader.max_wait_local:]:
                move_job_to_cluster(id_job)
        while len(scheduled_jobs_local) > 0 and nb_started < NB_RUN:
            start_job(scheduled_jobs_local.pop(0))
            nb_started += 1
        if config_reader.batch_system_type != "local" and len(scheduled_jobs_local) > config_reader.max_wait_local:
            for id_job in scheduled_jobs_local[config_reader.max_wait_local:]:
                move_job_to_cluster(id_job)
        for job in scheduled_jobs_cluster:
            start_job(job["job_id"], job["batch_type"])
        # Wait before return
        _printer("Sleeping...")
        time.sleep(15)
        _printer("\n")



import os
import sys
import re
import getpass
import psutil
from crontab import CronTab
from dgenies.config_reader import AppConfigReader


class Crons:

    """
    Manage crontab jobs (webserver mode)
    """

    def __init__(self, base_dir, debug):
        """

        :param base_dir: software base directory path
        :type base_dir: str
        :param debug: True to enable debug mode
        :type debug: bool
        """
        self.base_dir = base_dir
        self.debug = debug
        self.my_cron = CronTab(user=getpass.getuser())
        self.config = AppConfigReader()
        self.local_scheduler_pid_file = os.path.join(self.config.config_dir, ".local_scheduler_pid")

    def clear(self, kill_scheduler=True, remove_pid_file=True):
        """
        Clear all crons

        :param kill_scheduler: if True, kill local scheduler currently running
        :type kill_scheduler: bool
        :param remove_pid_file: if True, remove pid file if local scheduler was killed successfully
        :type remove_pid_file: bool
        """
        # Remove old crons:
        self.my_cron.remove_all(comment="dgenies")
        self.my_cron.write()
        if kill_scheduler:
            # Kill local scheduler:
            if os.path.exists(self.local_scheduler_pid_file):
                with open(self.local_scheduler_pid_file) as p_f:
                    pid = int(p_f.readline().strip("\n"))
                if psutil.pid_exists(pid) and remove_pid_file:
                    p = psutil.Process(pid)
                    p.terminate()
                    os.remove(self.local_scheduler_pid_file)

    def start_all(self):
        """
        Start all crons
        """
        self.clear(False)
        self.init_clean_cron()
        self.init_launch_local_cron()

    @staticmethod
    def _get_python_exec():
        """
        Get python executable path
        """
        pyexec = sys.executable
        match = re.match(r"^(.+)/lib/(python[^/]+)/((site-packages/bin/python)|())$", pyexec)
        if match:
            pyexec = "%s/bin/%s" % (match.group(1), match.group(2))
        return pyexec

    def init_clean_cron(self):
        """
        Initialize clean cron: will clear old jobs.
        Clean cron is launched at 1h00am each day
        """
        clean_time = self.config.cron_clean_time
        clean_freq = self.config.cron_clean_freq
        if self.base_dir is not None:
            job = self.my_cron.new(self._get_python_exec() +
                                   " {0}/bin/clean_jobs.py > {1}/clean.log 2>&1".format(self.base_dir,
                                                                                        self.config.log_dir),
                                   comment="dgenies")
            job.day.every(clean_freq)
            job.hour.on(clean_time[0])
            job.minute.on(clean_time[1])
            self.my_cron.write()
        else:
            raise Exception("Crons: base_dir must not be None")

    def init_launch_local_cron(self):
        """
        Try to launch local scheduler (if not already launched)
        """
        if self.base_dir is not None:
            pyexec = self._get_python_exec()
            logs = os.path.join(self.config.log_dir, "local_scheduler.log") if self.debug else "/dev/null"
            job = self.my_cron.new("{0}/bin/start_local_scheduler.sh {0} {1} {2} {3} > /dev/null 2>&1 &".
                                   format(self.base_dir, pyexec, self.local_scheduler_pid_file, logs),
                                   comment="dgenies")
            job.minute.every(1)
            self.my_cron.write()


from dgenies import app, app_title, app_folder, config_reader, mailer, APP_DATA, MODE, DEBUG, VERSION

import os
import sys
import time
import datetime
import shutil
import re
import threading
import traceback
import json
from flask import render_template, request, url_for, jsonify, Response, abort, send_file, Markup
from pathlib import Path
from dgenies.lib.paf import Paf
from dgenies.lib.job_manager import JobManager
from dgenies.lib.functions import Functions, ALLOWED_EXTENSIONS
from dgenies.lib.upload_file import UploadFile
from dgenies.lib.fasta import Fasta
from dgenies.lib.latest import Latest
from dgenies.tools import Tools
from markdown import Markdown
from markdown.extensions.toc import TocExtension
from markdown.extensions.tables import TableExtension
import tarfile
from xopen import xopen
from jinja2 import Environment
if MODE == "webserver":
    from dgenies.database import Session, Gallery
    from peewee import DoesNotExist


@app.context_processor
def global_templates_variables():
    """
    Global variables used for any view
    """
    return {
        "title": app_title,
        "mode": MODE,
        "all_jobs": Functions.get_list_all_jobs(MODE),
        "cookie_wall": config_reader.cookie_wall,
        "legal_pages": config_reader.legal,
        "debug": DEBUG
    }


# Main
@app.route("/", methods=['GET'])
def main():
    """
    Index page
    """
    if MODE == "webserver":
        pict = Gallery.select().order_by("id")
        if len(pict) > 0:
            pict = pict[0].picture
        else:
            pict = None
    else:
        pict = None
    return render_template("index.html", menu="index", pict=pict)


@app.route("/run", methods=['GET'])
def run():
    """
    Run page
    """
    inforun = None
    inforun_file = os.path.join(config_reader.config_dir, ".inforun")
    if os.path.exists(inforun_file):
        try:
            with open(inforun_file, "r") as info:
                inforun = json.loads(info.read())
        except json.JSONDecodeError:
            print("Unable to parse inforun file. Ignoring it.", file=sys.stderr)
            pass
    tools = Tools().tools
    tools_names = sorted(list(tools.keys()), key=lambda x: (tools[x].order, tools[x].name))
    tools_ava = {}
    tools_options = {}
    for tool_name, tool in tools.items():
        tools_ava[tool_name] = 1 if tool.all_vs_all is not None else 0
        tools_options[tool_name] = tool.options
    if MODE == "webserver":
        with Session.connect():
            s_id = Session.new()
    else:
        upload_folder = Functions.random_string(20)
        tmp_dir = config_reader.upload_folder
        upload_folder_path = os.path.join(tmp_dir, upload_folder)
        while os.path.exists(upload_folder_path):
            upload_folder = Functions.random_string(20)
            upload_folder_path = os.path.join(tmp_dir, upload_folder)
        s_id = upload_folder
    id_job = Functions.random_string(5) + "_" + datetime.datetime.fromtimestamp(time.time()).strftime('%Y%m%d%H%M%S')
    if "id_job" in request.args:
        id_job = request.args["id_job"]
    email = ""
    if "email" in request.args:
        email = request.args["email"]
    return render_template("run.html", id_job=id_job, email=email,
                           menu="run", allowed_ext=ALLOWED_EXTENSIONS, s_id=s_id,
                           max_upload_file_size=config_reader.max_upload_file_size,
                           example=config_reader.example_target != "",
                           target=os.path.basename(config_reader.example_target),
                           query=os.path.basename(config_reader.example_query), tools_names=tools_names, tools=tools,
                           tools_ava=tools_ava, tools_options=tools_options, version=VERSION, inforun=inforun)


@app.route("/run-test", methods=['GET'])
def run_test():
    """
    Run test page (used to simulate a real client run)
    """
    if MODE == "webserver":
        print(config_reader.allowed_ip_tests)
        if request.remote_addr not in config_reader.allowed_ip_tests:
            return abort(404)
        with Session.connect():
            return Session.new()
    return abort(500)


# Launch analysis
@app.route("/launch_analysis", methods=['POST'])
def launch_analysis():
    """
    Launch the job
    """
    if MODE == "webserver":
        try:
            with Session.connect():
                session = Session.get(s_id=request.form["s_id"])
        except DoesNotExist:
            return jsonify({"success": False, "errors": ["Session has expired. Please refresh the page and try again"]})
        upload_folder = session.upload_folder
        # Delete session:
        session.delete_instance()
    else:
        upload_folder = request.form["s_id"]

    id_job = request.form["id_job"]
    email = request.form["email"]
    file_query = request.form["query"]
    file_query_type = request.form["query_type"]
    file_target = request.form["target"]
    file_target_type = request.form["target_type"]
    tool = request.form["tool"] if "tool" in request.form else None
    tool_options = request.form.getlist("tool_options[]")
    alignfile = request.form["alignfile"] if "alignfile" in request.form else None
    alignfile_type = request.form["alignfile_type"] if "alignfile_type" in request.form else None
    backup = request.form["backup"] if "backup" in request.form else None
    backup_type = request.form["backup_type"] if "backup_type" in request.form else None

    # Check form:
    form_pass = True
    errors = []

    if alignfile is not None and alignfile_type is None:
        errors.append("Server error: no alignfile_type in form. Please contact the support")
        form_pass = False

    if backup is not None and backup != "" and (backup_type is None or backup_type == ""):
        errors.append("Server error: no backup_type in form. Please contact the support")
        form_pass = False

    if backup is not None and backup != "":
        alignfile = ""
        file_query = ""
        file_target = ""
    else:
        backup = None
        if file_target == "":
            errors.append("No target fasta selected")
            form_pass = False

    if tool is not None and tool not in Tools().tools:
        errors.append("Tool unavailable: %s" % tool)
        form_pass = False

    valid_options, options = get_tools_options(tool, tool_options)
    if not valid_options:
        errors.append("Chosen options unavailable")
        form_pass = False

    if id_job == "":
        errors.append("Id of job not given")
        form_pass = False

    if MODE == "webserver":
        if email == "":
            errors.append("Email not given")
            form_pass = False
        elif not re.match(r"^.+@.+\..+$", email):
            # The email regex is simple because checking email address is not simple (RFC3696).
            # Sending an email to the address is the most reliable way to check if the email address is correct.
            # The only constrains we set on the email address are:
            # - to have at least one @ in it, with something before and something after
            # - to have something.tdl syntax for email server, as it will be used over Internet (not mandatory in RFC)
            errors.append("Email is invalid")
            form_pass = False

    # Form pass
    if form_pass:
        # Get final job id:
        id_job = re.sub('[^A-Za-z0-9_\-]+', '', id_job.replace(" ", "_"))
        id_job_orig = id_job
        i = 2
        while os.path.exists(os.path.join(APP_DATA, id_job)):
            id_job = id_job_orig + ("_%d" % i)
            i += 1

        folder_files = os.path.join(APP_DATA, id_job)
        os.makedirs(folder_files)

        # Save files:
        query = None
        if file_query != "":
            example = False
            if file_query.startswith("example://"):
                example = True
                query_path = config_reader.example_query
                query_name = os.path.basename(query_path)
                file_query_type = "local"
            else:
                query_name = os.path.splitext(file_query.replace(".gz", ""))[0] if file_query_type == "local" else None
                if file_query_type == "local":
                    query_path = os.path.join(app.config["UPLOAD_FOLDER"], upload_folder, file_query)
                    if os.path.exists(query_path):
                        if " " in file_query:
                            new_query_path = os.path.join(app.config["UPLOAD_FOLDER"], upload_folder,
                                                                       file_query.replace(" ", "_"))
                            shutil.move(query_path, new_query_path)
                            query_path = new_query_path
                    else:
                        errors.append("Query file not correct!")
                        form_pass = False
                else:
                    query_path = file_query
            query = Fasta(name=query_name, path=query_path, type_f=file_query_type, example=example)
        example = False
        target = None
        if file_target != "":
            if file_target.startswith("example://"):
                example = True
                target_path = config_reader.example_target
                target_name = os.path.basename(target_path)
                file_target_type = "local"
            else:
                target_name = os.path.splitext(file_target.replace(".gz", ""))[0] if file_target_type == "local" \
                    else None
                if file_target_type == "local":
                    target_path = os.path.join(app.config["UPLOAD_FOLDER"], upload_folder, file_target)
                    if os.path.exists(target_path):
                        if " " in target_path:
                            new_target_path = os.path.join(app.config["UPLOAD_FOLDER"], upload_folder,
                                                           file_target.replace(" ", "_"))
                            shutil.move(target_path, new_target_path)
                            target_path = new_target_path
                    else:
                        errors.append("Target file not correct!")
                        form_pass = False
                else:
                    target_path = file_target
            target = Fasta(name=target_name, path=target_path, type_f=file_target_type, example=example)

        if alignfile is not None and alignfile != "" and backup is not None:
            Path(os.path.join(folder_files, ".align")).touch()

        align = None
        if alignfile is not None and alignfile != "":
            alignfile_name = os.path.splitext(alignfile)[0] if alignfile_type == "local" else None
            alignfile_path = os.path.join(app.config["UPLOAD_FOLDER"], upload_folder, alignfile) \
                if alignfile_type == "local" else alignfile
            if alignfile_type == "local" and not os.path.exists(alignfile_path):
                errors.append("Alignment file not correct!")
                form_pass = False
            align = Fasta(name=alignfile_name, path=alignfile_path, type_f=alignfile_type)

        bckp = None
        if backup is not None:
            backup_name = os.path.splitext(backup)[0] if backup_type == "local" else None
            backup_path = os.path.join(app.config["UPLOAD_FOLDER"], upload_folder, backup) \
                if backup_type == "local" else backup
            if backup_type == "local" and not os.path.exists(backup_path):
                errors.append("Backup file not correct!")
                form_pass = False
            bckp = Fasta(name=backup_name, path=backup_path, type_f=backup_type)

        if form_pass:
            # Launch job:
            job = JobManager(id_job=id_job,
                             email=email,
                             query=query,
                             target=target,
                             align=align,
                             backup=bckp,
                             mailer=mailer,
                             tool=tool,
                             options=options)
            if MODE == "webserver":
                job.launch()
            else:
                job.launch_standalone()
            return jsonify({"success": True, "redirect": url_for(".status", id_job=id_job)})
    if not form_pass:
        return jsonify({"success": False, "errors": errors})


# Status of a job
@app.route('/status/<id_job>', methods=['GET'])
def status(id_job):
    """
    Status page

    :param id_job: job id
    :type id_job: str
    """
    job = JobManager(id_job)
    j_status = job.status()
    mem_peak = j_status["mem_peak"] if "mem_peak" in j_status else None
    if mem_peak is not None:
        mem_peak = "%.1f G" % (mem_peak / 1024.0 / 1024.0)
    time_e = j_status["time_elapsed"] if "time_elapsed" in j_status else None
    if time_e is not None:
        if time_e < 60:
            time_e = "%d secs" % time_e
        else:
            minutes = time_e // 60
            seconds = time_e - minutes * 60
            time_e = "%d min %d secs" % (minutes, seconds)
    format = request.args.get("format")
    if format is not None and format == "json":
        return jsonify({
            "status": j_status["status"],
            "error": j_status["error"].replace("#ID#", ""),
            "id_job": id_job,
            "mem_peak": mem_peak,
            "time_elapsed": time_e
        })
    return render_template("status.html", status=j_status["status"],
                           error=j_status["error"].replace("#ID#", ""),
                           id_job=id_job, menu="results", mem_peak=mem_peak,
                           time_elapsed=time_e, do_align=job.do_align(),
                           query_filtered=job.is_query_filtered(), target_filtered=job.is_target_filtered())


# Results path
@app.route("/result/<id_res>", methods=['GET'])
def result(id_res):
    """
    Result page

    :param id_res: job id
    :type id_res: str
    """
    res_dir = os.path.join(APP_DATA, id_res)
    return render_template("result.html", id=id_res, menu="result", current_result=id_res,
                           is_gallery=Functions.is_in_gallery(id_res, MODE),
                           fasta_file=Functions.query_fasta_file_exists(res_dir))


@app.route("/gallery", methods=['GET'])
def gallery():
    """
    Gallery page
    """
    if MODE == "webserver":
        return render_template("gallery.html", items=Functions.get_gallery_items(), menu="gallery")
    return abort(404)


@app.route("/gallery/<filename>", methods=['GET'])
def gallery_file(filename):
    """
    Getting gallery illustration

    :param filename: filename of the PNG file
    """
    if MODE == "webserver":
        try:
            return send_file(os.path.join(config_reader.app_data, "gallery", filename))
        except FileNotFoundError:
            abort(404)
    return abort(404)


def get_tools_options(tool_name, chosen_options):
    """
    Transform options chosen in javascript into parameters

    :return: True is chosen options are valid + a string containing optional parameters to use with tool
    :rtype: boolean, str
    """
    tools = Tools().tools
    if tool_name is None:
        return True, None
    tool = tools[tool_name] if tool_name in tools else None
    tool_options = tool.options if tool is not None else None
    # We filter options for the chosen tool.
    tool_prefix = "tool-options-%s-" % tool_name
    filtered_options = [s[len(tool_prefix):].split("-") for s in chosen_options if s.startswith(tool_prefix)]
    valid = True
    options_params = None
    try:
        options_params = [tool_options[int(o)]['entries'][int(e)]['value'] for o, e in filtered_options]
    except KeyError:
        valid = False
    except IndexError:
        valid = False
    return True, " ".join(options_params)


def get_file(file, gzip=False):  # pragma: no cover
    """
    Download a file

    :param file: filename
    :type file: str
    :param gzip: is file gzipped?
    :type gzip: bool
    """
    try:
        # Figure out how flask returns static files
        # Tried:
        # - render_template
        # - send_file
        # This should not be so non-obvious
        return open(file, "rb" if gzip else "r").read()
    except FileNotFoundError:
        abort(404)
    except IOError as exc:
        print(exc.__traceback__)
        abort(500)


@app.route("/documentation/run", methods=['GET'])
def documentation_run():
    """
    Documentation run page
    """
    latest = Latest()
    version = latest.latest
    max_upload_file_size = config_reader.max_upload_file_size
    if max_upload_file_size == -1:
        max_upload_file_size = "no limit"
    else:
        max_upload_file_size = Functions.get_readable_size(max_upload_file_size, 0)
    max_upload_size = config_reader.max_upload_size
    if max_upload_size == -1:
        max_upload_size = "no limit"
    else:
        max_upload_size = Functions.get_readable_size(max_upload_size, 0)
    max_upload_size_ava = config_reader.max_upload_size_ava
    if max_upload_size_ava == -1:
        max_upload_size_ava = "no limit"
    else:
        max_upload_size_ava = Functions.get_readable_size(max_upload_size_ava, 0)
    with open(os.path.join(app_folder, "md", "doc_run.md"), "r",  encoding='utf-8') as install_instr:
        content = install_instr.read()
    env = Environment()
    template = env.from_string(content)
    content = template.render(mode=MODE, version=version, size=max_upload_file_size, size_unc=max_upload_size,
                              size_ava=max_upload_size_ava)
    md = Markdown(extensions=[TocExtension(baselevel=1)])
    content = Markup(md.convert(content))
    toc = Markup(md.toc)
    return render_template("documentation.html", menu="documentation", content=content, toc=toc)


@app.route("/documentation/result", methods=['GET'])
def documentation_result():
    """
    Documentation result page
    """
    with open(os.path.join(app_folder, "md", "user_manual.md"), "r",
              encoding='utf-8') as install_instr:
        content = install_instr.read()
    md = Markdown(extensions=[TocExtension(baselevel=1)])
    content = Markup(md.convert(content))
    toc = Markup(md.toc)
    return render_template("documentation.html", menu="documentation", content=content, toc=toc)


@app.route("/documentation/formats", methods=['GET'])
def documentation_formats():
    """
    Documentation formats page
    """
    with open(os.path.join(app_folder, "md", "doc_formats.md"), "r",
              encoding='utf-8') as install_instr:
        content = install_instr.read()
    md = Markdown(extensions=[TocExtension(baselevel=1), TableExtension()])
    content = Markup(md.convert(content))
    toc = Markup(md.toc)
    return render_template("documentation.html", menu="documentation", content=content, toc=toc)


@app.route("/documentation/dotplot", methods=['GET'])
def documentation_dotplot():
    """
    Documentation dotplot page
    """
    with open(os.path.join(app_folder, "md", "doc_dotplot.md"), "r",
              encoding='utf-8') as install_instr:
        content = install_instr.read()
    md = Markdown(extensions=[TocExtension(baselevel=1)])
    content = Markup(md.convert(content))
    toc = Markup(md.toc)
    return render_template("documentation.html", menu="documentation", content=content, toc=toc)


@app.route("/install", methods=['GET'])
def install():
    """
    Documentation: how to install? page
    """
    latest = Latest()

    with open(os.path.join(app_folder, "md", "INSTALL.md"), "r", encoding='utf-8') as install_instr:
        content = install_instr.read()
    env = Environment()
    template = env.from_string(content)
    content = template.render(version=latest.latest, win32=latest.win32)
    md = Markdown(extensions=[TocExtension(baselevel=1)])
    content = Markup(md.convert(content))
    toc = Markup(md.toc)
    return render_template("documentation.html", menu="install", content=content, toc=toc)


@app.route("/contact", methods=['GET'])
def contact():
    """
    Contact page
    """
    return render_template("contact.html", menu="contact")


@app.route("/legal/<page>", methods=['GET'])
def legal(page):
    """
    Display legal things
    """
    if page not in config_reader.legal:
        abort(404)
    with open(config_reader.legal[page], "r", encoding='utf-8') as md_page:
        content = md_page.read()
    env = Environment()
    template = env.from_string(content)
    content = template.render()
    md = Markdown(extensions=['extra', 'toc'])
    content = Markup(md.convert(content))
    return render_template("simple.html", menu="legal", content=content)


@app.route("/paf/<id_res>", methods=['GET'])
def download_paf(id_res):
    """
    Download PAF file of a job

    :param id_res: job id
    :type id_res: str
    """
    map_file = os.path.join(APP_DATA, id_res, "map.paf.sorted")
    if not os.path.exists(map_file):
        map_file = os.path.join(APP_DATA, id_res, "map.paf")
    if not os.path.exists(map_file):
        abort(404)
    content = get_file(map_file)
    return Response(content, mimetype="text/plain")


# Get graph (ajax request)
@app.route('/get_graph', methods=['POST'])
def get_graph():
    """
    Get dot plot data for a job
    """
    id_f = request.form["id"]
    paf = os.path.join(APP_DATA, id_f, "map.paf")
    idx1 = os.path.join(APP_DATA, id_f, "query.idx")
    idx2 = os.path.join(APP_DATA, id_f, "target.idx")

    paf = Paf(paf, idx1, idx2)

    if paf.parsed:
        valid = os.path.join(APP_DATA, id_f, ".valid")
        if not os.path.exists(valid):
            Path(valid).touch()
        res = paf.get_d3js_data()
        res["success"] = True
        return jsonify(res)
    return jsonify({"success": False, "message": paf.error})


@app.route('/sort/<id_res>', methods=['POST'])
def sort_graph(id_res):
    """
    Sort dot plot to reference

    :param id_res: job id
    :type id_res: str
    """
    if not os.path.exists(os.path.join(APP_DATA, id_res, ".all-vs-all")):
        paf_file = os.path.join(APP_DATA, id_res, "map.paf")
        idx1 = os.path.join(APP_DATA, id_res, "query.idx")
        idx2 = os.path.join(APP_DATA, id_res, "target.idx")
        paf = Paf(paf_file, idx1, idx2, False)
        paf.sort()
        if paf.parsed:
            res = paf.get_d3js_data()
            res["success"] = True
            return jsonify(res)
        return jsonify({"success": False, "message": paf.error})
    return jsonify({"success": False, "message": "Sort is not available for All-vs-All mode"})


@app.route('/reverse-contig/<id_res>', methods=['POST'])
def reverse_contig(id_res):
    """
    Reverse contig order

    :param id_res: job id
    :type id_res: str
    """
    contig_name = request.form["contig"]
    if not os.path.exists(os.path.join(APP_DATA, id_res, ".all-vs-all")):
        paf_file = os.path.join(APP_DATA, id_res, "map.paf")
        idx1 = os.path.join(APP_DATA, id_res, "query.idx")
        idx2 = os.path.join(APP_DATA, id_res, "target.idx")
        paf = Paf(paf_file, idx1, idx2, False)
        paf.reverse_contig(contig_name)
        if paf.parsed:
            res = paf.get_d3js_data()
            res["success"] = True
            return jsonify(res)
        return jsonify({"success": False, "message": paf.error})
    return jsonify({"success": False, "message": "Sort is not available for All-vs-All mode"})


@app.route('/freenoise/<id_res>', methods=['POST'])
def free_noise(id_res):
    """
    Remove noise from the dot plot

    :param id_res: job id
    :type id_res: str
    """
    paf_file = os.path.join(APP_DATA, id_res, "map.paf")
    idx1 = os.path.join(APP_DATA, id_res, "query.idx")
    idx2 = os.path.join(APP_DATA, id_res, "target.idx")
    paf = Paf(paf_file, idx1, idx2, False)
    paf.parse_paf(noise=request.form["noise"] == "1")
    if paf.parsed:
        res = paf.get_d3js_data()
        res["success"] = True
        return jsonify(res)
    return jsonify({"success": False, "message": paf.error})


@app.route('/get-fasta-query/<id_res>', methods=['POST'])
def build_fasta(id_res):
    """
    Generate the fasta file of query

    :param id_res: job id
    :type id_res: str
    """
    res_dir = os.path.join(APP_DATA, id_res)
    lock_query = os.path.join(res_dir, ".query-fasta-build")
    is_sorted = os.path.exists(os.path.join(res_dir, ".sorted"))
    compressed = request.form["gzip"].lower() == "true"
    query_fasta = Functions.get_fasta_file(res_dir, "query", is_sorted)
    if query_fasta is not None:
        if is_sorted and not query_fasta.endswith(".sorted"):
            # Do the sort
            Path(lock_query).touch()
            if not compressed or MODE == "standalone":  # If compressed, it will took a long time, so not wait
                Path(lock_query + ".pending").touch()
            index_file = os.path.join(res_dir, "query.idx.sorted")
            if MODE == "webserver":
                thread = threading.Timer(1, Functions.sort_fasta, kwargs={
                    "job_name": id_res,
                    "fasta_file": query_fasta,
                    "index_file": index_file,
                    "lock_file": lock_query,
                    "compress": compressed,
                    "mailer": mailer,
                    "mode": MODE
                })
                thread.start()
            else:
                Functions.sort_fasta(job_name=id_res,
                                     fasta_file=query_fasta,
                                     index_file=index_file,
                                     lock_file=lock_query,
                                     compress=compressed,
                                     mailer=None,
                                     mode=MODE)
            if not compressed or MODE == "standalone":
                if MODE == "webserver":
                    i = 0
                    time.sleep(5)
                    while os.path.exists(lock_query) and (i < 2 or MODE == "standalone"):
                        i += 1
                        time.sleep(5)
                os.remove(lock_query + ".pending")
                if os.path.exists(lock_query):
                    return jsonify({"success": True, "status": 1, "status_message": "In progress"})
                return jsonify({"success": True, "status": 2, "status_message": "Done",
                                "gzip": compressed})
            else:
                return jsonify({"success": True, "status": 1, "status_message": "In progress"})
        elif is_sorted and os.path.exists(lock_query):
            # Sort is already in progress
            return jsonify({"success": True, "status": 1, "status_message": "In progress"})
        else:
            # No sort to do or sort done
            if compressed and not query_fasta.endswith(".gz.fasta"):
                # If compressed file is asked, we must compress it now if not done before...
                Path(lock_query).touch()
                thread = threading.Timer(1, Functions.compress_and_send_mail, kwargs={
                    "job_name": id_res,
                    "fasta_file": query_fasta,
                    "index_file": os.path.join(res_dir, "query.idx.sorted"),
                    "lock_file": lock_query,
                    "mailer": mailer
                })
                thread.start()
                return jsonify({"success": True, "status": 1, "status_message": "In progress"})
            return jsonify({"success": True, "status": 2, "status_message": "Done",
                            "gzip": query_fasta.endswith(".gz") or query_fasta.endswith(".gz.sorted")})
    else:
        return jsonify({"success": False,
                        "message": "Unable to get fasta file for query. Please contact us to report the bug"})


def build_query_as_reference(id_res):
    """
    Build fasta of query with contigs order like reference

    :param id_res: job id
    :type id_res: str
    """
    paf_file = os.path.join(APP_DATA, id_res, "map.paf")
    idx1 = os.path.join(APP_DATA, id_res, "query.idx")
    idx2 = os.path.join(APP_DATA, id_res, "target.idx")
    paf = Paf(paf_file, idx1, idx2, False, mailer=mailer, id_job=id_res)
    paf.parse_paf(False, True)
    if MODE == "webserver":
        thread = threading.Timer(0, paf.build_query_chr_as_reference)
        thread.start()
        return True
    return paf.build_query_chr_as_reference()


@app.route('/build-query-as-reference/<id_res>', methods=['POST'])
def post_query_as_reference(id_res):
    """
    Launch build fasta of query with contigs order like reference

    :param id_res: job id
    :type id_res: str
    """
    build_query_as_reference(id_res)
    return jsonify({"success": True})


@app.route('/get-query-as-reference/<id_res>', methods=['GET'])
def get_query_as_reference(id_res):
    """
    Get fasta of query with contigs order like reference

    :param id_res: job id
    :type id_res: str
    """
    if MODE != "standalone":
        return abort(404)
    return send_file(build_query_as_reference(id_res))


@app.route('/download/<id_res>/<filename>')
def download_file(id_res, filename):
    """
    Download a file from a job

    :param id_res: job id
    :type id_res: str
    :param filename: file name
    :type filename: str
    """
    file_dl = os.path.join(APP_DATA, id_res, filename)
    if os.path.isfile(file_dl):
        return send_file(file_dl)
    return abort(404)


@app.route('/fasta-query/<id_res>', defaults={'filename': ""}, methods=['GET'])
@app.route('/fasta-query/<id_res>/<filename>', methods=['GET'])  # Use fake URL in mail to set download file name
def dl_fasta(id_res, filename):
    """
    Download fasta file

    :param id_res: job id
    :type id_res: str
    :param filename: file name (not used, but can be in the URL to define download filename to the browser)
    :type filename: str
    """
    res_dir = os.path.join(APP_DATA, id_res)
    lock_query = os.path.join(res_dir, ".query-fasta-build")
    is_sorted = os.path.exists(os.path.join(res_dir, ".sorted"))
    if not os.path.exists(lock_query) or not is_sorted:
        query_fasta = Functions.get_fasta_file(res_dir, "query", is_sorted)
        if query_fasta is not None:
            if query_fasta.endswith(".gz") or query_fasta.endswith(".gz.sorted"):
                content = get_file(query_fasta, True)
                return Response(content, mimetype="application/gzip")
            content = get_file(query_fasta)
            return Response(content, mimetype="text/plain")
    abort(404)


@app.route('/qt-assoc/<id_res>', methods=['GET'])
def qt_assoc(id_res):
    """
    Query - Target association TSV file

    :param id_res:
    :return:
    """
    res_dir = os.path.join(APP_DATA, id_res)
    if os.path.exists(res_dir) and os.path.isdir(res_dir):
        paf_file = os.path.join(APP_DATA, id_res, "map.paf")
        idx1 = os.path.join(APP_DATA, id_res, "query.idx")
        idx2 = os.path.join(APP_DATA, id_res, "target.idx")
        try:
            paf = Paf(paf_file, idx1, idx2, False)
            paf.parse_paf(False)
        except FileNotFoundError:
            print("Unable to load data!")
            abort(404)
            return False
        csv_content = paf.build_query_on_target_association_file()
        return Response(csv_content, mimetype="text/plain")
    abort(404)


@app.route('/no-assoc/<id_res>', methods=['POST'])
def no_assoc(id_res):
    """
    Get contigs that match with None target

    :param id_res: job id
    :type id_res: str
    """
    res_dir = os.path.join(APP_DATA, id_res)
    if os.path.exists(res_dir) and os.path.isdir(res_dir):
        paf_file = os.path.join(APP_DATA, id_res, "map.paf")
        idx1 = os.path.join(APP_DATA, id_res, "query.idx")
        idx2 = os.path.join(APP_DATA, id_res, "target.idx")
        try:
            paf = Paf(paf_file, idx1, idx2, False)
        except FileNotFoundError:
            print("Unable to load data!")
            abort(404)
            return False
        file_content = paf.build_list_no_assoc(request.form["to"])
        empty = file_content == "\n"
        return jsonify({
            "file_content": file_content,
            "empty": empty
        })
    abort(404)


@app.route('/summary/<id_res>', methods=['POST'])
def summary(id_res):
    """
    Get Dot plot summary data

    :param id_res: job id
    :type id_res: str
    """
    paf_file = os.path.join(APP_DATA, id_res, "map.paf")
    idx1 = os.path.join(APP_DATA, id_res, "query.idx")
    idx2 = os.path.join(APP_DATA, id_res, "target.idx")
    try:
        paf = Paf(paf_file, idx1, idx2, False)
    except FileNotFoundError:
        return jsonify({
            "success": False,
            "message": "Unable to load data!"
        })
    percents = None
    s_status = "waiting"  # Accepted values: waiting, done, fail
    status_file = os.path.join(APP_DATA, id_res, ".summarize")
    fail_file = status_file + ".fail"
    if not os.path.exists(status_file):  # The job is finished or not started
        if not os.path.exists(fail_file):  # The job has not started yet or has successfully ended
            percents = paf.get_summary_stats()
            if percents is None:  # The job has not started yet
                Path(status_file).touch()
                thread = threading.Timer(0, paf.build_summary_stats, kwargs={"status_file": status_file})
                thread.start()
            else:  # The job has successfully ended
                s_status = "done"
        else:  # The job has failed
            s_status = "fail"

    if s_status == "waiting":  # The job is running
        # Check if the job end in the next 30 seconds
        nb_iter = 0
        while os.path.exists(status_file) and not os.path.exists(fail_file) and nb_iter < 10:
            time.sleep(3)
            nb_iter += 1
        if not os.path.exists(status_file):  # The job has ended
            percents = paf.get_summary_stats()
            if percents is None:  # The job has failed
                s_status = "fail"
            else:  # The job has successfully ended
                s_status = "done"

    if s_status == "fail":
        return jsonify({
            "success": False,
            "message": "Build of summary failed. Please contact us to report the bug"
        })
    return jsonify({
        "success": True,
        "percents": percents,
        "status": s_status
    })


@app.route('/backup/<id_res>')
def get_backup_file(id_res):
    """
    Download archive backup file of a job

    :param id_res: job id
    :type id_res: str
    """
    res_dir = os.path.join(APP_DATA, id_res)
    tar = os.path.join(res_dir, "%s.tar.gz" % id_res)
    with xopen(tar, mode="wb", compresslevel=9) as gz_file:
        with tarfile.open(fileobj=gz_file, mode="w|") as tarf:
            for file in ("map.paf", "target.idx", "query.idx"):
                tarf.add(os.path.join(res_dir, file), arcname=file)
    return send_file(tar)


def get_filter_out(id_res, type_f):
    """
    Download filter fasta, when it has been filtered before job run

    :param id_res: job id
    :type id_res: str
    :param type_f: type of fasta (query or target)
    :type type_f: str
    """
    filter_file = os.path.join(APP_DATA, id_res, ".filter-" + type_f)
    return Response(get_file(filter_file), mimetype="text/plain")


@app.route('/filter-out/<id_res>/query')
def get_filter_out_query(id_res):
    """
    Download query filtered fasta, when it has been filtered before job run

    :param id_res: job id
    :type id_res: str
    """
    return get_filter_out(id_res=id_res, type_f="query")


@app.route('/filter-out/<id_res>/target')
def get_filter_out_target(id_res):
    """
    Download target filtered fasta, when it has been filtered before job run

    :param id_res: job id
    :type id_res: str
    """
    return get_filter_out(id_res=id_res, type_f="target")


@app.route('/viewer/<id_res>')
def get_viewer_html(id_res):
    """
    Get HTML file with offline interactive viewer inside

    :param id_res: job id
    :type id_res: str
    """
    paf = os.path.join(APP_DATA, id_res, "map.paf")
    idx1 = os.path.join(APP_DATA, id_res, "query.idx")
    idx2 = os.path.join(APP_DATA, id_res, "target.idx")

    paf = Paf(paf, idx1, idx2)

    if paf.parsed:
        res = paf.get_d3js_data()
        res["success"] = True
        percents = paf.get_summary_stats()
        with open(os.path.join(app_folder, "static", "js", "dgenies-offline-result.min.js"), "r") as js_min:
            js = js_min.read()
        with open(os.path.join(app_folder, "static", "css", "dgenies-offline-result.min.css"), "r") as css_min:
            css = css_min.read()
        return render_template("map_offline.html", json=json.dumps(res), version=VERSION, js=js, css=css,
                               percents=percents)
    return abort(403)


@app.route("/ask-upload", methods=['POST'])
def ask_upload():
    """
    Ask for upload: to keep a max number of concurrent uploads
    """
    if MODE == "standalone":
        return jsonify({
            "success": True,
            "allowed": True
        })
    try:
        s_id = request.form['s_id']
        with Session.connect():
            session = Session.get(s_id=s_id)
            allowed = session.ask_for_upload(True)
        return jsonify({
            "success": True,
            "allowed": allowed
        })
    except DoesNotExist:
        return jsonify({"success": False, "message": "Session not initialized. Please refresh the page."})


@app.route("/ping-upload", methods=['POST'])
def ping_upload():
    """
    When upload waiting, ping to be kept in the waiting line
    """
    if MODE == "webserver":
        s_id = request.form['s_id']
        with Session.connect():
            session = Session.get(s_id=s_id)
            session.ping()
    return "OK"


@app.route("/upload", methods=['POST'])
def upload():
    """
    Do upload of a file
    """
    try:
        s_id = request.form['s_id']
        if MODE == "webserver":
            try:
                with Session.connect():
                    session = Session.get(s_id=s_id)
                    if session.ask_for_upload(False):
                        folder = session.upload_folder
                    else:
                        return jsonify({"files": [], "success": "ERR", "message": "Not allowed to upload!"})
            except DoesNotExist:
                return jsonify(
                    {"files": [], "success": "ERR", "message": "Session not initialized. Please refresh the page."})
        else:
            folder = s_id

        files = request.files[list(request.files.keys())[0]]

        if files:
            filename = files.filename
            folder_files = os.path.join(app.config["UPLOAD_FOLDER"], folder)
            if not os.path.exists(folder_files):
                os.makedirs(folder_files)
            filename = Functions.get_valid_uploaded_filename(filename, folder_files)
            mime_type = files.content_type

            if not Functions.allowed_file(files.filename, request.form['formats'].split(",")):
                result = UploadFile(name=filename, type_f=mime_type, size=0, not_allowed_msg="File type not allowed")
                shutil.rmtree(folder_files)

            else:
                # save file to disk
                uploaded_file_path = os.path.join(folder_files, filename)
                files.save(uploaded_file_path)

                # get file size after saving
                size = os.path.getsize(uploaded_file_path)

                # return json for js call back
                result = UploadFile(name=filename, type_f=mime_type, size=size)

            return jsonify({"files": [result.get_file()], "success": "OK"})

        return jsonify({"files": [], "success": "404", "message": "No file provided"})
    except:  # Except all possible exceptions to prevent crashes
        traceback.print_exc()
        return jsonify({"files": [], "success": "ERR", "message": "An unexpected error has occurred on upload. "
                                                                  "Please contact the support."})


@app.route("/send-mail/<id_res>", methods=['POST'])
def send_mail(id_res):
    """
    Send mail

    :param id_res: job id
    :type id_res: str
    """
    allowed = False
    key_file = None
    if "key" in request.form:
        key = request.form["key"]
        res_dir = os.path.join(APP_DATA, id_res)
        key_file = os.path.join(res_dir, ".key")
        if os.path.exists(key_file):
            with open(key_file) as k_f:
                true_key = k_f.readline().strip("\n")
                allowed = key == true_key
    if allowed:
        os.remove(key_file)
        job_mng = JobManager(id_job=id_res, mailer=mailer)
        job_mng.set_inputs_from_res_dir()
        job_mng.send_mail()
        return "OK"
    else:
        abort(403)


@app.route("/delete/<id_res>", methods=['POST'])
def delete_job(id_res):
    """
    Delete a job

    :param id_res: job id
    :type id_res: str
    """
    job = JobManager(id_job=id_res)
    success, error = job.delete()
    return jsonify({
        "success": success,
        "error": error
    })


#!/usr/bin/env python3

import argparse
import time
import os
import sys
from split_fa import Splitter
from filter_contigs import Filter
from index import index_file


def index_fasta(name, filepath, out_dir, type_f, dofilter = True):
    """
    Index and filter fasta

    :param name: name of the specie
    :param filepath: full path of the fasta file
    :param out_dir: output folder
    :param type_f: type of fasta (query or target)
    """
    uncompressed = None
    if filepath.endswith(".gz") and dofilter:
        uncompressed = filepath[:-3]
    index = os.path.join(out_dir, type_f + ".idx")
    success, nb_contigs, error = index_file(filepath, name, index, uncompressed)
    if success:
        is_filtered = False
        if dofilter:
            in_fasta = filepath
            if uncompressed is not None:
                in_fasta = uncompressed
            filtered_fasta = os.path.join(os.path.dirname(in_fasta), "filtered_" + os.path.basename(in_fasta))
            filter_f = Filter(fasta=in_fasta,
                              index_file=index,
                              type_f=type_f,
                              min_filtered=nb_contigs / 4,
                              split=False,
                              out_fasta=filtered_fasta,
                              replace_fa=True)
            is_filtered = filter_f.filter()
        if uncompressed is not None:
            if is_filtered:
                os.remove(filepath)
                with open(os.path.join(out_dir, "." + type_f), "w") as save_file:
                    save_file.write(uncompressed)
            else:
                os.remove(uncompressed)

    else:
        print("###ERR### %s fasta  file is not valid:<br/> %s" % (type_f[0].upper() + type_f[1:], error), file=sys.stderr)
        if uncompressed is not None:
            try:
                os.remove(uncompressed)
            except FileNotFoundError:
                pass
        exit(1)


parser = argparse.ArgumentParser(description="Split huge contigs")
parser.add_argument('-q', '--query', type=str, required=False, help="Query fasta file")
parser.add_argument('-u', '--query-split', type=str, required=False, help="Query fasta file split")
parser.add_argument('-t', '--target', type=str, required=False, help="Target fasta file")
parser.add_argument('-n', '--query-name', type=str, required=False, help="Query name")
parser.add_argument('-m', '--target-name', type=str, required=False, help="Target name")
parser.add_argument('-s', '--size', type=int, required=False, default=10,
                    help="Max size of contigs (Mb) - for query split")
parser.add_argument('-p', '--preptime-file', type=str, required=True, help="File into save prep times")
parser.add_argument('--split', type=bool, const=True, nargs="?", required=False, default=False,
                        help="Split query")
parser.add_argument('--index-only', type=bool, const=True, nargs="?", required=False, default=False,
                        help="Index files only. No split, no filter.")
args = parser.parse_args()

if args.index_only and args.split:
    raise Exception("--index-only and --split arguments are mutually exclusive")

out_dir = os.path.dirname(args.target)

with open(args.preptime_file, "w") as ptime:
    ptime.write(str(round(time.time())) + "\n")
    if args.query is not None:
        if args.split:
            print("Splitting query...")
            fasta_in = args.query
            index_split = os.path.join(out_dir, "query_split.idx")
            splitter = Splitter(input_f=fasta_in, name_f=args.query_name, output_f=args.query_split,
                                query_index=index_split)
            success, error = splitter.split()
            if success:
                filtered_fasta = os.path.join(os.path.dirname(args.query_split), "filtered_" +
                                              os.path.basename(args.query_split))
                filter_f = Filter(fasta=args.query_split,
                                  index_file=index_split,
                                  type_f="query",
                                  min_filtered=splitter.nb_contigs / 4,
                                  split=True,
                                  out_fasta=filtered_fasta,
                                  replace_fa=True)
                filter_f.filter()
            else:
                print("###ERR### Query fasta file is not valid:<br/> %s" % error, file=sys.stderr)
                exit(1)
        else:
            print("Indexing query...")
            index_fasta(name=args.query_name, filepath=args.query, out_dir=out_dir, type_f="query",
                        dofilter=not args.index_only)
    if args.target is not None:
        print("Indexing target...")
        index_fasta(name=args.target_name, filepath=args.target, out_dir=out_dir, type_f="target",
                    dofilter=not args.index_only)

    ptime.write(str(round(time.time())) + "\n")

print("DONE!")



import os
import sys
import re
import inspect
from pathlib import Path
from configparser import RawConfigParser, NoOptionError, NoSectionError
from dgenies.lib.decorators import Singleton


@Singleton
class AppConfigReader:
    """
    Store all configs
    """

    def __init__(self):
        """
        All "get_*" functions results are stored in the "self.*" corresponding attribute
        Example: results of the get_upload_folder function is stored in self.upload_folder
        """
        self.app_dir = os.path.dirname(inspect.getfile(self.__class__))
        config_file = []
        config_file_search = [os.path.join(os.path.abspath(os.sep), "dgenies", "application.properties"),
                              "/etc/dgenies/application.properties",
                              "/etc/dgenies/application.properties.local",
                              os.path.join(str(Path.home()), ".dgenies", "application.properties"),
                              os.path.join(str(Path.home()), ".dgenies", "application.properties.local"),
                              os.path.join(self.app_dir, '..', 'etc', 'dgenies', 'application.properties')]

        if os.name == "nt":
            config_file.insert(1, os.path.join(sys.executable, '..', "application.properties"))
            config_file.insert(1, os.path.join(sys.executable, '..', "application.properties.local"))

        for my_config_file in config_file_search:
            if os.path.exists(my_config_file):
                config_file.append(my_config_file)

        config_file.append(os.path.join(self.app_dir, "application-dev.properties"))
        config_file.append(os.path.join(self.app_dir, "application-dev.properties.local"))

        if len(config_file) == 0:
            raise FileNotFoundError("ERROR: application.properties not found.")
        self.reader = RawConfigParser()
        self.reader.read(config_file)
        for attr in dir(self):
            attr_o = getattr(self, attr)
            if attr.startswith("_get_") and callable(attr_o):
                try:
                    setattr(self, attr[5:], attr_o())
                except Exception as e:
                    print(e)

    def _replace_vars(self, path, config=False):
        new_path = path.replace("###USER###", os.path.expanduser("~"))\
            .replace("###PROGRAM###", os.path.dirname(os.path.dirname(os.path.realpath(__file__))))\
            .replace("###SYSEXEC###", os.path.dirname(sys.executable))
        if "###CONFIG###" in new_path:
            if config:
                raise Exception("###CONFIG### tag not allowed for config dir")
            else:
                return new_path.replace("###CONFIG###", self._get_config_dir())
        return new_path

    def _get_config_dir(self):
        try:
            return self._replace_vars(self.reader.get("global", "config_dir"), True)
        except NoOptionError:
            return self._replace_vars("###USER###/.dgenies")

    def _get_upload_folder(self):
        try:
            return self._replace_vars(self.reader.get("global", "upload_folder"))
        except NoOptionError:
            raise Exception("No upload folder found in application.properties (global section)")

    def _get_app_data(self):
        try:
            return self._replace_vars(self.reader.get("global", "data_folder"))
        except NoOptionError:
            raise Exception("No data folder found in application.properties (global section)")

    def _get_batch_system_type(self):
        try:
            return self.reader.get("global", "batch_system_type")
        except NoOptionError:
            return "local"

    def _get_web_url(self):
        try:
            return self._replace_vars(self.reader.get("global", "web_url"))
        except NoOptionError:
            return "http://localhost:5000"

    def _get_max_upload_size(self):
        try:
            max_size_b = self._replace_vars(self.reader.get("global", "max_upload_size"))
            if max_size_b == "-1":
                return -1
            size_v = float(max_size_b[:-1])
            size_unit = max_size_b[-1].upper()
            if size_unit not in ["M", "G"]:
                raise ValueError("Max size unit must be M or G")
            max_size = int(size_v * 1024 * 1024)
            if size_unit == "G":
                max_size *= 1024
            return max_size
        except NoOptionError:
            return -1

    def _get_max_upload_size_ava(self):
        try:
            max_size_b = self._replace_vars(self.reader.get("global", "max_upload_size_ava"))
            if max_size_b == "-1":
                return -1
            size_v = float(max_size_b[:-1])
            size_unit = max_size_b[-1].upper()
            if size_unit not in ["M", "G"]:
                raise ValueError("Max size unit must be M or G")
            max_size = int(size_v * 1024 * 1024)
            if size_unit == "G":
                max_size *= 1024
            return max_size
        except NoOptionError:
            return -1

    def _get_max_upload_file_size(self):
        try:
            max_size_b = self._replace_vars(self.reader.get("global", "max_upload_file_size"))
            if max_size_b == "-1":
                return -1
            size_v = float(max_size_b[:-1])
            size_unit = max_size_b[-1].upper()
            if size_unit not in ["M", "G"]:
                raise ValueError("Max size unit must be M or G")
            max_size = int(size_v * 1024 * 1024)
            if size_unit == "G":
                max_size *= 1024
            return max_size
        except NoOptionError:
            return 1024 * 1024 * 1024

    def _get_max_nb_lines(self):
        try:
            return int(self._replace_vars(self.reader.get("global", "max_nb_lines")))
        except NoOptionError:
            return 100000

    def _get_database_type(self):
        try:
            return self.reader.get("database", "type")
        except (NoSectionError, NoOptionError):
            return "sqlite"

    def _get_database_url(self):
        try:
            url = self._replace_vars(self.reader.get("database", "url"))
            if self._get_database_type() == "sqlite" and url != ":memory:":
                parent_dir = os.path.dirname(url)
                if not os.path.exists(parent_dir):
                    try:
                        os.makedirs(parent_dir)
                    except FileNotFoundError:
                        pass
            return url
        except (NoSectionError, NoOptionError):
            return self._replace_vars("###USER###/.dgenies/database.sqlite")

    def _get_database_port(self):
        try:
            return int(self.reader.get("database", "port"))
        except (NoSectionError, NoOptionError, ValueError):
            db_type = self._get_database_type()
            if db_type == "mysql":
                return 3306
            elif db_type == "sqlite":
                return -1
            raise Exception("Missing parameter: database port")

    def _get_database_db(self):
        try:
            db = self.reader.get("database", "db")
            if db == "":
                raise ValueError()
            return db
        except (NoSectionError, NoOptionError, ValueError):
            if self._get_database_type() == "sqlite":
                return ""
            raise Exception("Missing parameter: database db name")

    def _get_database_user(self):
        try:
            user = self.reader.get("database", "user")
            if user == "":
                raise ValueError()
            return user
        except (NoSectionError, NoOptionError, ValueError):
            if self._get_database_type() == "sqlite":
                return ""
            raise Exception("Missing parameter: database user")

    def _get_database_password(self):
        try:
            passwd = self.reader.get("database", "password")
            if passwd == "":
                raise ValueError()
            return passwd
        except (NoSectionError, NoOptionError, ValueError):
            if self._get_database_type() == "sqlite":
                return ""
            raise Exception("Missing parameter: database password")

    def _get_mail_status_sender(self):
        try:
            return self._replace_vars(self.reader.get("mail", "status"))
        except (NoSectionError, NoOptionError):
            return "status@dgenies"

    def _get_mail_reply(self):
        try:
            return self._replace_vars(self.reader.get("mail", "reply"))
        except (NoSectionError, NoOptionError):
            return "status@dgenies"

    def _get_mail_org(self):
        try:
            return self._replace_vars(self.reader.get("mail", "org"))
        except (NoSectionError, NoOptionError):
            return None

    def _get_send_mail_status(self):
        try:
            return self.reader.get("mail", "send_mail_status").lower() == "true"
        except (NoSectionError, NoOptionError):
            return True

    def _get_disable_mail(self):
        try:
            return self.reader.get("mail", "disable").lower() == "true"
        except (NoSectionError, NoOptionError):
            return False

    def _get_cron_clean_time(self):
        try:
            value = self.reader.get("cron", "clean_time").lower()
            match = re.match(r"(([0-9])|([0-1][0-9])|(2[0-3]))[hH]([0-5][0-9])", value)
            if match is not None:
                return [int(match.group(1)), int(match.group(5))]
            else:
                print("Incorrect clean hour format!")
                return [1, 0]
        except (NoOptionError, NoSectionError):
            return [1, 0]

    def _get_cron_clean_freq(self):
        try:
            return int(self.reader.get("cron", "clean_freq"))
        except (NoOptionError, NoSectionError):
            return 1

    def _get_local_nb_runs(self):
        try:
            return int(self.reader.get("jobs", "run_local"))
        except (NoOptionError, NoSectionError):
            return 1

    def _get_nb_data_prepare(self):
        try:
            return int(self.reader.get("jobs", "data_prepare"))
        except (NoOptionError, NoSectionError):
            return 2

    def _get_max_concurrent_dl(self):
        try:
            return int(self.reader.get("jobs", "max_concurrent_dl"))
        except (NoOptionError, NoSectionError):
            return 5

    def _get_drmaa_lib_path(self):
        try:
            path = self.reader.get("cluster", "drmaa_lib_path")
            if path != "###SET_IT###":
                return path
            return None
        except (NoOptionError, NoSectionError):
            if self._get_batch_system_type() != "local":
                raise Exception("No drmaa library set. It is required if the batch system type is not 'local'")
            return None

    def _get_drmaa_native_specs(self):
        try:
            return self.reader.get("cluster", "native_specs")
        except (NoOptionError, NoSectionError):
            return "###DEFAULT###"

    def _get_max_run_local(self):
        try:
            return int(self.reader.get("cluster", "max_run_local"))
        except (NoOptionError, NoSectionError):
            return 10

    def _get_max_wait_local(self):
        try:
            return int(self.reader.get("cluster", "max_wait_local"))
        except (NoOptionError, NoSectionError):
            return 5

    def _get_min_query_size(self):
        try:
            size_b = self.reader.get("cluster", "min_query_size")
            size_v = int(size_b[:-1])
            size_unit = size_b[-1].upper()
            if size_unit not in ["M", "G"]:
                raise ValueError("Min query size unit must be M or G")
            min_size = size_v * 1024 * 1024
            if size_unit == "G":
                min_size *= 1024
            return min_size
        except (NoOptionError, NoSectionError):
            return 0

    def _get_min_target_size(self):
        try:
            size_b = self.reader.get("cluster", "min_target_size")
            size_v = int(size_b[:-1])
            size_unit = size_b[-1].upper()
            if size_unit not in ["M", "G"]:
                raise ValueError("Min query size unit must be M or G")
            min_size = size_v * 1024 * 1024
            if size_unit == "G":
                min_size *= 1024
            return min_size
        except (NoOptionError, NoSectionError):
            return 0

    def _get_cluster_prepare_script(self):
        try:
            return self._replace_vars(self.reader.get("cluster", "prepare_script"))
        except (NoOptionError, NoSectionError):
            return self._replace_vars("###PROGRAM###/bin/all_prepare.py")

    def _get_cluster_python_exec(self):
        try:
            return self._replace_vars(self.reader.get("cluster", "python3_exec"))
        except (NoOptionError, NoSectionError):
            return "python3"

    def _get_cluster_memory(self):
        try:
            memory = int(self.reader.get("cluster", "memory"))
            return memory
        except (NoOptionError, NoSectionError):
            return 32

    def _get_cluster_memory_ava(self):
        try:
            memory = int(self.reader.get("cluster", "memory_ava"))
            return memory
        except (NoOptionError, NoSectionError):
            return self._get_cluster_memory()

    def _get_debug(self):
        try:
            return self.reader.get("debug", "enable").lower() == "true"
        except (NoOptionError, NoSectionError):
            return False

    def _get_log_dir(self):
        try:
            log_dir = self._replace_vars(self.reader.get("debug", "log_dir"))
        except (NoOptionError, NoSectionError):
            log_dir = self._replace_vars("###CONFIG###/logs")
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        elif not os.path.isdir(log_dir):
            raise TypeError("Log dir must be a directory")
        return log_dir

    def _get_allowed_ip_tests(self):
        allowed_ip = {"127.0.0.1"}
        try:
            allowed_ip_txt = self.reader.get("debug", "allowed_ip_tests")
            for ip in re.split(r",(\s+)?", allowed_ip_txt):
                allowed_ip.add(ip)
        except (NoOptionError, NoSectionError):
            pass
        return allowed_ip

    def _get_example_query(self):
        try:
            return self.reader.get("example", "query")
        except (NoOptionError, NoSectionError):
            return ""

    def _get_example_target(self):
        try:
            return self.reader.get("example", "target")
        except (NoOptionError, NoSectionError):
            return ""

    def _get_analytics_enabled(self):
        try:
            return self.reader.get("analytics", "enable_logging_runs").lower() == "true"
        except (NoOptionError, NoSectionError):
            return False

    def _get_anonymous_analytics(self):
        try:
            return not self.reader.get("analytics", "disable_anonymous_analytics").lower() == "true"
        except (NoOptionError, NoSectionError):
            return True

    def _get_analytics_groups(self):
        try:
            return [(option, self.reader.get("analytics_groups", option)) for option in self.reader.options("analytics_groups")]
        except (NoOptionError, NoSectionError):
            return []

    def _get_cookie_wall(self):
        try:
            return self.reader.get("legal", "cookie_wall")
        except (NoOptionError, NoSectionError):
            return None

    def _get_legal(self):
        try:
            return {option: self.reader.get("legal", option) for option in self.reader.options("legal")
                    if option not in ["cookie_wall"]}
        except (NoOptionError, NoSectionError):
            return {}
